#!/usr/bin/perl

#############################################################################
# fwrules.pl: Analyze and print a report of Firewall-1 rules and objects
#             Note: unsupported by Checkpoint or representatives.
#
# Copyright (C) 2001 Volker Tanger
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# For bug reports and suggestions or if you just want to talk to me please
# contact me at volker.tanger@wyae.de
#
# Updates will be available at  http://www.wyae.de/software/
# please check there for updates prior to submitting patches!
#
# For list of changes please refer to the HISTORY file. Thanks.
#############################################################################


#
# Used modules
#
use Getopt::Long;

$VERSION	 = '7.2.0';
$SCR_NAME	 = 'FW1rules';
$HTMLtitle	 = 'FireWall-1 Rule Base and Object Definitions';

$FW1rules	 = 'Standard.W';
$FW1objects	 = 'objects.C';
$FW1user	 = 'fwauth.NDB';
$IconPathName	 = 'icons';

$LogFile	 = 'FWrules.log';
$TemplateINCLUDE = '';
$TemplateITEMSEP = ' ';

%colors =
  ('black'              => '#111111',
   'blue'               => '#3333FF',
   'cyan'               => '#00FFFF',
   'dark green'         => '#006600',
   'dark orchid'        => '#CC33CC',
   'firebrick'          => '#993333',
   'foreground'         => '#111111',
   'forest green'       => '#66FF99',
   'gold'               => '#FFCC33',
   'green'              => '#00FF00',
   'magenta'            => '#FF99FF',
   'medium slate blue'  => '#6666FF',
   'navy blue'          => '#000077',
   'orange'             => '#FF9900',
   'red'                => '#FF0000',
   'sienna'             => '#996600',
   'yellow'             => '#FFFF00');


%netmasktranslation = (
	'255.255.255.255'  => '32',
	'255.255.255.254'  => '31',
	'255.255.255.252'  => '30',
	'255.255.255.248'  => '29',
	'255.255.255.240'  => '28',
	'255.255.255.224'  => '27',
	'255.255.255.192'  => '26',
	'255.255.255.128'  => '25',
	'255.255.255.0'	   => '24',
	'255.255.254.0'	   => '23',
	'255.255.252.0'	   => '22',
	'255.255.248.0'	   => '21',
	'255.255.240.0'	   => '20',
	'255.255.224.0'	   => '19',
	'255.255.192.0'	   => '18',
	'255.255.128.0'	   => '17',
	'255.255.0.0'	   => '16',
	'255.254.0.0'	   => '15',
	'255.252.0.0'	   => '14',
	'255.248.0.0'	   => '13',
	'255.240.0.0'	   => '12',
	'255.224.0.0'	   => '11',
	'255.192.0.0'	   => '10',
	'255.128.0.0'	   => '9',
	'255.0.0.0'	   => '8',
	'254.0.0.0'	   => '7',
	'252.0.0.0'	   => '6',
	'248.0.0.0'	   => '5',
	'240.0.0.0'	   => '4',
	'224.0.0.0'	   => '3',
	'192.0.0.0'	   => '2',
	'128.0.0.0'	   => '1',
	'0.0.0.0'	   => '0'  );

%NATtranslation = (
	0 => 'hide',
	1 => 'static' );

$SynDefender[0] = 'None';
$SynDefender[1] = 'SYN Relay';
$SynDefender[2] = 'SYN Gateway';
$SynDefender[3] = 'Passive SYN Gateway';

##########################################################################
# print out Usage
sub Usage{
  print STDERR "

   fw1rules.pl
	[--objects=<objects file>]
	[--rules=<rules file>]
	[--users=<users file>]
	[--all_objects] [--all_services] [--all_users]
	[--no_rules] [--no_objects] [--no_services] [--no_users] 
	[--with_implicit_rules] [--with_ip] [--unused_only]
	[--sort_by_type]
	[--verbose] [--version]
	
	[--dump_nat_tsv=<tab separated NAT rules file>] 
	[--dump_nat_txt=<NAT rules text file>] 
	[--dump_objects_tsv=<tab separated objects file>] 
	[--dump_objects_txt=<objects text file>] 
	[--dump_rules_tsv=<tab separated rules file>] 
	[--dump_rules_txt=<rules text file>] 
	[--dump_services_tsv=<tab separated services file>] 
	[--dump_services_txt=<services text file>]
	[--dump_users_tsv=<tab separated users file>] 
	[--dump_users_txt=<users text file>]
	[--dump_properties=<text file>] 
	[--dump_unused_objects=<text file>] 

	[--use-css=<css file>] 
	[--with_colors] [--icon_path=<path>]
	[--title=<document title>]
	[--output_html=<html file>]
	
	[--template=<template file>]
	[--template_include=<template include filename>]
	[--output=<filled template file>]




Optional Parameters:
--------------------

   --rules=<rule file>: Location of FireWall-1 rule file. 
	Default is 'Standard.W'

   --objects=<objects file>: Location of FireWall-1 objects file.
	  Default is 'objects.C'

   --users=<users file>: Location of FireWall-1 user file.
	 Default is 'fwauth.NDB'

   --all_objects  /  --all_services  /  --all_users: 
	it will list ALL objects/services/users. By default only those 
	in use will be listed.

   --no_rules  /  --no_objects  /  --no_services  /  --no_users: 
	it will NOT read nor list rules/objects/services/users. 
	Default disabled.
	--no_rules implies --all_objs and --all_services

   --with_implicit_rules: include the implicit rules into the tables
	(Note: this does not affect the additional dumps - see below)

   --with_ip: prints IP addresses along with the object names into the
	access rule & objects dump for HTML.
	(Note: this does not affect the additional dumps - see below)

   --sort_by_type: use object and service type as primary sorting key 
   	to group them. Secondary key is alphabetically then. Default
	sorting is just plain alphabetically (case ignore).

   --with_colors: Add html color tags to network and service objects
	to reflect the colors used in the FW-1 GUI.  By default, this
	is OFF to enhance readability (e.g. on printouts).

   --icon_path=<path>: Location of the icon graphic files. 
   	Defaults is './icons'

   --use-css=<css file>: include reference to the CSS file given (for
	nicer output)

   --title=<title>: Here you can set a custom document title.

   --verbose: prints debugging information to STDERR and FWrules.log
   
   --version: prints version and exists


To give an overall documentation of the ruleset (and used objects) you
can print the output into different file formats. Other formats will
be supported as soon as someone supplies according documentation or
patch. You can call any number of these options (okay, only one per
file type per program run):

	[--output_html=<html file>]


To fill a custom template with the current firewall configuration you
have to give a template file and the name of the resulting file. You
only can give one template/output pair per program run. 

A number of predefined templates is supplied with the tarball in the
subdirectory templates/

A special case is the option --template_include where one can set the
template variable include to be replaced with the command line
parameter. This comes in handy when defining files to be included e.g.
for producing complete documentations - see the templates/configdoc.tex 
template. 

	[--template=<template file>]
	[--template_include=<template include filename>]
	[--output=<filled template file>]


To additionally dump objects, service definitions or rules into
TabSeparated or TXT files use the commands below. This comes in handy
if used for further processing or inclusion into documentation. You
can call any number of these options (okay, only one per file type):

   --dump_nat_tsv=<tab separated NAT rules file>
   --dump_nat_txt=<NAT rules text file>
   --dump_objects_tsv=<tab separated objects file>
   --dump_objects_txt=<objects text file>
   --dump_rules_tsv=<tab separated rules file>
   --dump_rules_txt=<rules text file>
   --dump_services_tsv=<tab separated services file>
   --dump_services_txt=<services text file>
   --dump_unused_objects=<services text file>
   --dump_properties=<properties text file>

\n";
}

##########################################################################
# correct Micro$oft stuff (line end, spaces)
sub fromdos {
    $line = $_[0];
    $line =~ s/\n//g;
    $line =~ s/\r//g;
    $line =~ s/        /\t/g;
    return $line;
}

##########################################################################
# print out comments / errors
sub PrintLog{
    my ($msg) = $_[0];
    if ($FLAG_verbose){
	print STDERR "$msg";
	print LOGFILE "$msg";
    }
}

##########################################################################
# print only if second parameter not empty 
sub PrintNonempty{ 
    my ($FILE)   = $_[0];
    my ($first)  = $_[1];
    my ($second) = $_[2];
    
    if ( "$second" ne '' ) {
	printf $FILE ("$first","$second");
    }
}

##########################################################################
# read all network entities/objects defined
#
# Object variables where obj_name equals the hash for each of these:
#
#	$obj_number 	= number of objects
#	@obj_name 	= names of all objects
#	%obj_type 	= host, network, gateway, group
#	%obj_location 	= 0=internal, 1=external
#	%obj_is_fw1 	= has FW1 installed? 0=false, 1=true
#	%obj_ipaddr 	= IP Address
#	%obj_netmask	= netmask
#	%obj_NATadr 	= NAT address for implicit NAT
#	%obj_NATtype 	= 0=hide, 1=static
#	%obj_members 	= members, if a group
#	%obj_comment 	= comment for the object
#	%obj_colour 	= colour the object is to be displayed with 
#	%obj_used 	= wether the object is used in the rulbase
#			  (set later when evaluating the ruleset)

sub ReadNetworkObjects{
    my ($dummy)     = '';
    my ($name)      = '';
    my ($lineparam) = '';
    my ($amember)   = '';
    my ($members)   = '';

    $obj_number = 0;
    while ( ($line = <INFILE> ) && ( fromdos("$line") ne "\t)" ) ) {
          fromdos($line);
          while ( $line !~ /\t\t\: \(/ )  {
                $line = <INFILE>;
                fromdos($line);
          }
          ($dummy,$name) = split(/\(/,$line,2) ;
          $amember = '';
          $members = '';
          while ( ($line = <INFILE>) && ( fromdos("$line") ne "\t\t)" ) ) {
                fromdos($line);
                ($dummy,$lineparam) = split(/\(/,$line,2) ;
                $lineparam =~ s/\)$//;
		if ( $line =~ /^\t\t\t:type \(/ ){
		   $obj_type{$name} = lc($lineparam);
                } elsif ( $line =~ /^\t\t\t:location \(/ ){
		   $obj_location{$name} = ("$lineparam" eq 'external') * 1;
                } elsif ( $line =~ /^\t\t\t:firewall \(/ ){
		   $obj_is_fw1{$name} = ("$lineparam" eq 'installed') * 1;
                } elsif ( $line =~ /^\t\t\t:ipaddr \(/ ){
 		   $obj_ipaddr{$name} = $lineparam;
                } elsif ( $line =~ /^\t\t\t:netmask \(/ ){
                   $obj_netmask{$name} = $lineparam;
                } elsif ( $line =~ /^\t\t\t:valid_ipaddr \(/ ){
 		   $obj_NATadr{$name} = $lineparam;
                } elsif ( $line =~ /^\t\t\t:netobj_adtr_method \(/ ){
		   $obj_NATtype{$name} = ("$lineparam" eq 'adtr_static') * 1;
                } elsif ( $line =~ /^\t\t\t:comments \(/ ){
 		   $obj_comment{$name} = $lineparam;
		   $obj_comment{$name} =~ s/^\"|\"$//g;		#--- remove " at beginning and end
		   $obj_comment{$name} =~ s/;/ /g;
                } elsif ( $line =~ /^\t\t\t:color \(/ ){
 		   $obj_colour{$name} = lc($lineparam);
                } elsif ( $line =~ /^\t\t\t: / ){
                   ($dummy,$amember) = split(/:/,$line,2) ;
                   $members = "$members$amember";
                }
          }
          if ( "$obj_type{$name}" eq 'group' ) {
             ($dummy,$members) = split (/ /, $members, 2);
             $obj_members{$name} = $members;
          }
	  $obj_name[$obj_number] = $name;
	  $obj_number += 1;
          &PrintLog('.');
    }
    $obj_type{'Any'} = 'any';
    if ($FLAG_sortbytype) {
    	@obj_name = sort { $obj_type{"$a"} cmp $obj_type{"$b"} or lc($a) cmp lc($b) } @obj_name;
    } else {
    	@obj_name = sort { lc($a) cmp lc($b) } @obj_name;
    }
}

##########################################################################
# read all network services defined
#
# service variables where svc_name equals the hash for each of these:
#
#	$svc_number 	= number of services read
#	@svc_name 	= names of all services
#	%svc_type 	= tcp, udp, icmp, rpc, group
#	%svc_dst_port 	= destination port
#	%svc_src_low 	= range source port from
#	%svc_src_high 	= range source port to
#	%svc_match	= if MATCH defines (for RPCs)
#	%svc_prolog	= RPC prolog
#	%svc_members 	= members, if a group
#	%svc_comment 	= comment for the service
#	%svc_colour 	= colour of the service
#	%svc_used 	= wether the service is used in the rulbase
#			  (set later when evaluating the ruleset)
sub ReadServices{
    my ($dummy)    = '';
    my ($name)     = '';
    my ($amember)  = '';
    my ($members)  = '';

    while ( ($line = <INFILE>) && ( fromdos("$line") ne "\t)" ) ) {
          fromdos($line);
          while ( $line !~ /\t\t\: \(/ )  {
                $line = <INFILE>;
                fromdos($line);
          }
          ($dummy,$name) = split(/\(/,$line,2) ;
          $amember  = '';
          $members  = '';
          while ( ($line = <INFILE>) && ( fromdos("$line") ne "\t\t)" ) ) {
                fromdos($line);
                ($dummy,$lineparam) = split(/\(/,$line,2) ;
                $lineparam =~ s/\)$//;
                if ("$lineparam" =~ /"\>/){  # this stands for ports bigger than...
                   ($dummy,$lineparam) = split(/\>/,$lineparam,2) ;
                   ($lineparam,$dummy) = split(/"/,$lineparam,2) ;
                   $lineparam = "$lineparam:";
                }
                if ( $line =~ /^\t\t\t:type \(/ ){
                   $svc_type{$name} = lc($lineparam);
                } elsif ( $line =~ /^\t\t\t:exp \(/ ){           # ICMP extensions
                   $lineparam =~ s/\"//g;
                   $svc_dst_port{$name} = $ICMPtranslate{$lineparam};
                } elsif ( $line =~ /^\t\t\t:port \(/ ){          # TCP/UDP destination port
                   $lineparam =~ tr/-/:/;
                   $svc_dst_port{$name} = $lineparam;
                } elsif ( $line =~ /^\t\t\t:src_port_from \(/ ){
                   $svc_src_low{$name} = $lineparam;
                } elsif ( $line =~ /^\t\t\t:src_port_to \(/ ){
                   $svc_src_high{$name} = $lineparam;
                } elsif ( $line =~ /^\t\t\t:prematch \(/ ){
                   $svc_match{$name} = $lineparam;
                } elsif ( $line =~ /^\t\t\t:prolog \(/ ){
                   $svc_prolog{$name} = $lineparam;
                } elsif ( $line =~ /^\t\t\t:comments \(/ ){
                   $svc_comment{$name} = $lineparam;
		   $svc_comment{$name} =~ s/^\"|\"$//g;		#--- remove " at beginning and end
		   $svc_comment{$name} =~ s/;/ /g;
		} elsif ( $line =~ /^\t\t\t:color \(/ ){
                   $svc_colour{$name} = lc($lineparam);
                } elsif ( $line =~ /^\t\t\t: / ){
                   ($dummy,$amember) = split(/:\x20/,$line,2) ;
                   $members = "$members$amember";
                }
          }
	  $svc_name[$svc_number] = $name;
          if ( "$svc_type{$name}" eq 'group' ) {
                ($dummy,$members) = split (//, $members, 2);
                $svc_members{$name} = $members;
           }
           &PrintLog('.');
	   $svc_number += 1;
    }
    &PrintLog ("\n");
    $svc_type{'Any'} = 'any';
    if ($FLAG_sortbytype) {
    	@svc_name = sort { $svc_type{"$a"} cmp $svc_type{"$b"} or lc($a) cmp lc($b) } @svc_name;
    } else {
    	@svc_name = sort { lc($a) cmp lc($b) } @svc_name;
    }
}

##########################################################################
# read properties
#
#	%prop_setting{'XXX'}	= setting for XXX
#
#	of interest with respect to implicit rules:
#		rip, domain_udp, domain_tcp, established,
#		    icmpenable, fw1enable ==  true / false
#		rip_p, domain_udp_p, domain_tcp_p, established_p,
#		    icmpenable_p, fw1enable_p ==  first / "before last" / last
#
sub ReadProperties{
    my($line) = '';
    my($par)  = '';
    my($set)  = '';
    my($rest) = '';
    
    while ( ($line = <INFILE>) && ( fromdos("$line") ne "\t)" ) ) {
          $line = fromdos($line);
          &PrintLog('.');
          if ( "$line" =~ m/\t\t:.* \(.*\)$/ ){
	     ($par,$set) = split(/ \(/, $line, 2);
 	     $par =~ s/^\s+://;					#--- remove "    :" at the beginning
 	     $set =~ s/\)$//;					#--- remove ) at the end
 	     $set =~ s/^\"|\"$//g;				#--- remove " at beginning and end
             $prop_setting{"$par"} = "$set";
          }
    }
    &PrintLog("\n");
}


##########################################################################
# read users
#
# user variables where user_name equals the hash for each of these:
#
#	$user_number	= number of users read
#	@user_name	= names of all users
#	%user_type	= usrgroup, template, user
#	%user_members	= members, if a group
#	%user_comment	= comment for the user
#	%user_colour	= colour of the user
#	%user_used	= wether the user is used in the rulbase
#	                  (set later when evaluating the ruleset)
sub ReadUser {
    my ($i,$sh,$sl,$bh,$bl,$tmp,$StartBlk);
    my ($NumBlocks,@block,$ilen,$clen,$glen);
    my ($type,$st,$nh,$nl,$ihh,$ih,$il,$chh,$ch,$cl);
    my ($Users)        ='';
    # open input and seek to start of data
    open (INFILE,"$FW1user") 
	or die "Cannot open the userDB file $FW1user!\n\n";

    if(!seek(INFILE, 12288, 0)){
	&PrintLog("file to short(1)\n");
	return;
    }

    # read header
    if(read(INFILE, $tmp, 256) < 256) {
	&Printlog("file to short(2)\n");
	return;
    }
    #structur: ?,?,startblock,number of blocks
    ($sh,$sl,$bh,$bl)=unpack("x2CCCC", $tmp);
    $StartBlk=$sh*256+$sl;
    $NumBlocks=$bh*256+$bl;

    # read NumBlocks in array
    $i=1;
    while(read(INFILE, $tmp, 256)){
	$block[$i]=$tmp;
	$i++
    }

    # parse blocks and write records in variable
    $Users="(\n\t";
    for($i=1;$i<=$NumBlocks;$i++){
    ($st,$nh,$nl)=unpack("C3", $block[$i]);
    $next=($st&127)*65536+$nh*256+$nl;
    if($st & 128){    # first bit in block marks a new record
	($st,$nh,$nl,$ihh,$ih,$il,$chh,$ch,$cl,$type)=unpack("C3x3C2C3C3C", $block[$i]);
	$ilen=$ih*256+$il; # len of record name
	$clen=$chh*65536+$ch*256+$cl; # len of record content
	$glen=$clen+$ilen; # len of whole record
	$llen=$clen; # num of bytes to read
	if($glen>243){
	 $llen=243-$ilen;
	 $rlen=$clen-$llen;
	} else {
	 if($next) {
	   print STDERR "error 1 in Block $i\n";
	   $next=0;
	 }
	}
	$index=unpack("x13 a$ilen", $block[$i]);
	$cont=unpack("x13 x$ilen a$llen", $block[$i]);
	$content=": $cont";
	# going through the chain until end
	while($next){
	 $llen=$rlen;
	 if($rlen>253){
	   $llen=253;
	   $rlen-=253;
	 }
	 ($st,$nh,$nl,$cont)=unpack("C3a$llen", $block[$next]);
	 $curr=$next;
	 $next=($st&127)*65536+$nh*256+$nl;
	 if($next && ($llen < 253)) {
	   $next=0;
	   print STDERR "error 2 in Block $curr ($next:$llen)\n";
	 }
	 $content.=$cont;
	}
	$content=~s/\n/\n\t/g;
	$Users.="$content";
	&PrintLog('.');
    }
    }
    $Users.=")\n";

    my ($dummy)    = '';
    my ($name)     = '';
    my ($amember)  = '';
    my ($members)  = '';
    my ($user_number) = 0;

    @U=split(/\n/,$Users);
    while($line = shift(@U)) {
	 while ( $line !~ /\t:\s+\(/ && @U)  {
	       $line = shift(@U);
	       &PrintLog('-');
	 }
	 ($dummy,$name) = split(/\(/,$line,2) ;
	 $name =~ s/^"//;
	 $name =~ s/"$//;
	 $amember  = '';
	 $members  = '';
	 while ( ($line = shift(@U)) && ( $line !~ /^\t\)/ ) ) {
	       ($dummy,$lineparam) = split(/\(/,$line,2) ;
	       ($lineparam,$dummy) = split(/\)/,$lineparam,2) ;
	       if ( $line =~ /^\t\t:type \(/ ){
	          $user_type{$name} = lc($lineparam);
	       } elsif ( $line =~ /^\t\t:comments \(/ ){
	          $user_comment{$name} = $lineparam;
		  $user_comment{$name} =~ s/^\"|\"$//g;		#--- remove " at beginning and end
		  $user_comment{$name} =~ s/;/ /g;
	       } elsif ( $line =~ /^\t\t:fromhour \(/ ){
	          $user_from{$name} = lc($lineparam);
	       } elsif ( $line =~ /^\t\t:tohour \(/ ){
	          $user_to{$name} = lc($lineparam);
	       } elsif ( $line =~ /^\t\t:expiration_date \(/ ){
	          $user_expiration{$name} = lc($lineparam);
	       } elsif ( $line =~ /^\t\t:auth_method \(/ ){
	          $user_auth{$name} = lc($lineparam);
	       } elsif ( $line =~ /^\t\t:color \(/ ){
	          $user_colour{$name} = lc($lineparam);
	       } elsif ( $line =~ /^\t\t: / ){
	          ($dummy,$amember) = split(/: \(/,$line,2) ;
	          $amember =~ s/^"//;
	          $amember =~ s/"$//;
	          $members = "$members $amember";
	       }
	 }
	 $user_name[$user_number] = $name;
	 if ( "$user_type{$name}" eq 'usrgroup' ) {
	       $member =~ s/^ *//,$member;
	       $user_members{$name} = $members;
	  }
	  &PrintLog('.');
	  $user_number += 1;
    }
    &PrintLog ("\n");
    @user_name = sort (@user_name);
    if ($FLAG_sortbytype) {
	@user_name = sort { $user_type{"$a"} cmp $user_type{"$b"} or lc($a) cmp lc($b) } @user_name;
    } else {
	@user_name = sort { lc($a) cmp lc($b) } @user_name;
    }

    close (INFILE);
}

##########################################################################
# recursively set usage on objects
sub SetObjUsed{
    my ($index) = $_[0];
    my ($single) = '';
    my (@members);

    $obj_used{"$index"} += 1;		
    if ( "$obj_type{$index}" eq 'group'){
    	@members = split (/ /,$obj_members{$index});
	foreach $single (@members) {
	    &SetObjUsed("$single");
	}
    }
}


##########################################################################
# recursively set usage on services
sub SetSvcUsed{
    my ($index) = $_[0];
    my ($single) = '';
    my (@members);
    
    $svc_used{"$index"} += 1;		
    if ( "$svc_type{$index}" eq 'group'){
    	@members = split (//,$svc_members{$index});
	foreach $single (@members) {
	    &SetSvcUsed("$single");
	}
    }
}


##########################################################################
# recursively set usage on users
sub SetUserUsed{
    my ($index) = $_[0];
    my ($single) = '';
    my (@members);
  
    $user_used{"$index"} += 1;                
    if ( "$user_type{$index}" eq 'usrgroup'){
	   @members = split (/ /,$user_members{$index});
	foreach $single (@members) {
	   &SetUserUsed("$single");
	}
    }
}


##########################################################################
# read NAT rules
#
#	$nat_number	 	= number of NAT rules read (array starting at zero)
#	@nat_orig_from 		= original source object
#	@nat_orig_to 		= original destination object
#	@nat_orig_svc 		= original service object
#	@nat_transl_from 	= translated source object
#	@nat_transl_from_methd 	= translated source object method: 0=hide, 1=static
#	@nat_transl_to 		= translated destination object
#	@nat_transl_to_methd	= translated destination object method: 0=hide, 1=static
#	@nat_transl_svc 	= translated service object
#	@nat_transl_svc_methd 	= translated service object method: 0=hide, 1=static
#	@nat_install_on		= install rule on...
#
sub ReadNATrules{
    my ($mode)    = 'none';
    my ($param)   = '';
    my ($dummy)   = '';
    my ($wert)    = '';
    my ($user)    = '';
    my ($fileEOF) = 1;

    $nat_number = -1;
    while ( ( $line !~ /^\t\:filename \(/ ) && $fileEOF ) {
	 fromdos($line);
	 $mode    = 'none';
	 if ( $line !~ /^\t:rule_adtr \(/ ) {
	       &PrintLog('.');
	 } else {
	    $nat_number  += 1;
	    &PrintLog("\n\trule_adtr($nat_number)");

	    while ( ($line = <INFILE>) && ( fromdos("$line") ne "\t)" ) ) {
	       fromdos($line);
	       &PrintLog('.');
	       if ( $line =~ /^\t\t:comments \(/ ){
	               ($dummy,$wert) = split(/\(/,$line,2) ;
		       $wert =~ s/\)$//;			#--- remove ) at the end
		       $wert =~ s/^\"|\"$//g;			#--- remove " at beginning and end
		       $wert =~ s/;/ /g;
		       $nat_comment[$nat_number] = $wert ;
	       } elsif ( $line =~ /^\t\t:disabled \(true\)/ ){
	               $nat_disabled[$nat_number] = 1;
	       } elsif ( $line =~ /^\t\t:(src_adtr|dst_adtr|services_adtr|src_adtr_translated|dst_adtr_translated|services_adtr_translated|install) \(/ ){
	               ($dummy,$wert) = split(/:/,$line,2) ;
	               ($mode,$dummy) = split(/ /,$wert,2) ;
	       } elsif ("$mode" eq 'src_adtr') {
	           if ($line =~ /^\t\t\t: /) {
	               ($dummy,$wert) = split(/: /,$line,2) ;
	               $wert=~s/^\(//;
	               $nat_orig_from[$nat_number] = "$wert";
	               &SetObjUsed("$wert");
	           }
	       } elsif ("$mode" eq 'dst_adtr') {
	           if ($line =~ /^\t\t\t: /) {
	               ($dummy,$wert) = split(/: /,$line,2) ;
	               $wert=~s/^\(//;
	               $nat_orig_to[$nat_number] = "$wert";
	               &SetObjUsed("$wert");
	           }
	       } elsif ("$mode" eq 'services_adtr') {
	           if ($line =~ /^\t\t\t: /) {
	               ($dummy,$wert) = split(/: /,$line,2) ;
	               $wert=~s/^\(//;
	               $nat_orig_svc[$nat_number] = "$wert";
	               &SetSvcUsed("$wert");
	           }
	       } elsif ("$mode" eq 'src_adtr_translated') {
	           if ($line =~ /^\t\t\t: /) {
	               ($dummy,$wert) = split(/: /,$line,2) ;
	               $wert=~s/^\(//;
		       if ( lc("$wert") eq 'any' ) { $wert = 'Original'; }
	               $nat_transl_from[$nat_number] = "$wert";
	               &SetObjUsed("$wert");
	           } elsif ($line =~ /^\t\t\t:adtr_method/) {
	               if ( $line =~ m/adtr_method_static/ ) {
	                    $nat_transl_from_methd[$nat_number] = 1;
	               } else {
	                    $nat_transl_from_methd[$nat_number] = 0;
	               }
	           }
	       } elsif ("$mode" eq 'dst_adtr_translated') {
	           if ($line =~ /^\t\t\t: /) {
	               ($dummy,$wert) = split(/: /,$line,2) ;
	               $wert=~s/^\(//;
		       if ( lc("$wert") eq 'any' ) { $wert = 'Original'; }
	               $nat_transl_to[$nat_number] = "$wert";
	               &SetObjUsed("$wert");
	               $nat_transl_to_methd[$nat_number] = 1;
	           }
	       } elsif ("$mode" eq 'services_adtr_translated') {
	           if ($line =~ /^\t\t\t: /) {
	               ($dummy,$wert) = split(/: /,$line,2) ;
	               $wert=~s/^\(//;
		       if ( lc("$wert") eq 'any' ) { $wert = 'Original'; }
	               $nat_transl_svc[$nat_number] = "$wert";
	               &SetSvcUsed("$wert");
	               $nat_transl_svc_methd[$nat_number] = 1;
	           }
	       } elsif ("$mode" eq 'install') {
	           if  ( $line =~ /^\t\t\t: / ) {
	               ($dummy,$wert) = split(/: /,$line,2) ;
	               $wert=~s/^\(//;
	               if ( "$nat_install_on[$nat_number]" eq '') {
	                   $nat_install_on[$nat_number] = "$wert";
	               } else {        
	                   $nat_install_on[$nat_number] = "$nat_install_on[$nat_number] $wert";
	               }
	           }
	        }
	    } #--- inner while
	 } #--- else (i.e. rule processing)
	 $fileEOF = ($line = <INFILE>);
    } #--- outer while
}


##########################################################################
# read Access rules
#
#	$access_number	 	= number of access rules read (array starting at zero)
#	@access_disabled	= rule enabled=0, rule disabled=1
#	@access_from		= list of source objects, separated by space
#	@access_from_negated	= from-list negated=1, standard=0
#	@access_to		= list of destination objects, separated by space
#	@access_to_negated	= to-list negated=1, standard=0
#	@access_services	= list of services, separated by space
#	@access_services_negated= services-list negated=1, standard=0
#	@access_action		= action: deny, allow, encrypt, ...
#	@access_track		= log: long, short, account, ...
#	@access_time		= time object (not really implemented yet)
#	@access_install_on	= install rule on...
#	@access_comment		= comment on this particular rule
#
sub ReadAccessRules{
    my ($mode)    = 'none';
    my ($param)   = '';
    my ($dummy)   = '';
    my ($wert)    = '';
    my ($fileEOF) = 1;

    $access_number = -1;
    while ( ( $line !~ /^\t\:rule_adtr \(/ ) && $fileEOF ) {
          fromdos($line);
          $mode    = 'none';
          if ( $line !~ /^\t:rule \(/ ) {
                &PrintLog('.');
          } else {
             $access_number  += 1;
             &PrintLog("\n\trule($access_number)");

             while ( ($line = <INFILE>) && ( fromdos("$line") ne "\t)" ) ) {
		fromdos($line);
		&PrintLog('.');
		if ( $line =~ /^\t\t:comments \(/ ){
			($dummy,$wert) = split(/\(/,$line,2) ;
			$wert =~ s/\)$//;			#--- remove ) at the end
			$wert =~ s/^\"|\"$//g;			#--- remove " at beginning and end
			$wert =~ s/;/ /g;
			$access_comment[$access_number] = $wert ;
		} elsif ( $line =~ /^\t\t:disabled \(true\)/ ){
			$access_disabled[$access_number] = 1;
		} elsif ( $line =~ /^\t\t:(src|dst|services|action|track|install|time) \(/ ){
			($dummy,$wert) = split(/:/,$line,2) ;
			($mode,$dummy) = split(/ /,$wert,2) ;
		} elsif ("$mode" eq 'src') {
		      $access_from_negated[$access_number] = 0;
                      if ($line =~ /^\t\t\t: /) {
			($dummy,$wert) = split(/:\s+\(*/,$line,2) ;
			$wert=~s/^\(//;				#--- remove ( at the beginning of user-rules
			if ( "$access_from[$access_number]" eq '') {
			    $access_from[$access_number] = "$wert";
			} else {	
			    $access_from[$access_number] = "$access_from[$access_number]$wert";
			}
			&SetObjUsed("$wert");
                      } elsif ($line =~ /^\t\t\t:op \(\"not in\"\)/) {
			$access_from_negated[$access_number] = 1;
                      }
		} elsif ("$mode" eq 'dst') {
		      $access_to_negated[$access_number] = 0;
                      if ($line =~ /^\t\t\t: /) {
			($dummy,$wert) = split(/:\s+\(*/,$line,2) ;
			if ( "$access_to[$access_number]" eq '') {
			    $access_to[$access_number] = "$wert";
			} else {
			    $access_to[$access_number] = "$access_to[$access_number]$wert";
			}
			&SetObjUsed("$wert");
                      } elsif ($line =~ /^\t\t\t:op \(\"not in\"\)/) {
			$access_to_negated[$access_number] = 1;
                      }
		} elsif ("$mode" eq 'services') {
		      $access_services_negated[$access_number] = 0;
		      ($dummy,$wert) = split(/\(\"/,$line,2) ;   #--- just for security servers
		      ($wert,$dummy) = split(/\"/,$wert,2) ;
                      if ($line =~ /^\t\t\t: \(\"smtp-\>/) {
	 		if ( "$access_services[$access_number]" eq '') {
			    $access_services[$access_number] = "$wert";
			} else {
			    $access_services[$access_number] = "$access_services[$access_number]$wert";
			}
                      } elsif ($line =~ /^\t\t\t: \(\"(http-\>.*)\"/) {
                        if ( "$access_services[$access_number]" eq '') {
			    $access_services[$access_number] = "$wert";
			} else {
			    $access_services[$access_number] = "$access_services[$access_number]$wert";
			}
                     } elsif ($line =~ /^\t\t\t: \(\"ftp-\>/) {
	 		if ( "$access_services[$access_number]" eq '') {
			    $access_services[$access_number] = "$wert";
			} else {
			    $access_services[$access_number] = "$access_services[$access_number]$wert";
			}
                      } elsif ($line =~ /^\t\t\t:\s+/) {		# PS	(Any trouble
			($dummy,$wert) = split(/:\s+\(*/,$line,2) ;	# PS	(Any trouble
			if ( "$access_services[$access_number]" eq '') {
			    $access_services[$access_number] = "$wert";
			} else {
			    $access_services[$access_number] = "$access_services[$access_number]$wert";
			}
			&SetSvcUsed("$wert");
                      } elsif ($line =~ /^\t\t\t:op \(\"not in\"\)/) {
			$access_services_negated[$access_number] = 1;
                      }
		} elsif ("$mode" eq 'action') {
		      $wert =~ s/[\s\"]//g;		# PS
		      $wert = lc ($wert);		# PS
		      if  ( $line =~ /^\t\t\t:\s+\([a-z]*/ ) {	# PS
			($dummy,$wert) = split(/:\s+\(/,$line,2);	# PS
		     	$wert =~ s/[\s\"]//g;		# PS 
		      	$wert = lc ($wert);		# PS 
			$access_action[$access_number] = $wert;
                      }
		} elsif ("$mode" eq 'track') {
		      if  ( $line =~ /^\t\t\t: [A-Z]([a-z]*)/ ) {
			($dummy,$wert) = split(/: /,$line,2) ;
			$access_track[$access_number] = "$wert";
                      }
		} elsif ("$mode" eq 'install') {
		      if  ( $line =~ /^\t\t\t: / ) {
 			($dummy,$wert) = split(/:\s+\(*/,$line,2) ;		# PS left parenthesis
			$wert =~ s/\(//; # Handle Gateway object
			&SetObjUsed("$wert");
			if ( "$access_install_on[$access_number]" eq '') {
			    $access_install_on[$access_number] = "$wert";
			} else {	
			    $access_install_on[$access_number] = "$access_install_on[$access_number] $wert";
			}
                      }
		} elsif ("$mode" eq 'time') {
		      if  ( $line =~ /^\t\t\t: .*/ ) {
			($dummy,$wert) = split(/:\s+\(*/,$line,2) ;
			if ( "$access_time[$access_number]" eq '') {
			    $access_time[$access_number] = "$wert";
			} else {	
			    $access_time[$access_number] = "$access_time[$access_number] $wert";
			}
                      }
	 	}
	     } #--- inner while
	  } #--- else (i.e. rule processing)
	  $fileEOF = ($line = <INFILE>);
    } #--- outer while
}



##########################################################################
##########################################################################
###   Dump into single files
##########################################################################
##########################################################################


##########################################################################
# dump properties into a TXT file
#	filename for the dump

sub DumpProperties{
    my ($filename) 	= $_[0];

    open (DUMPFILE,">$filename") ;
    print DUMPFILE "Security Policy\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "Apply Gateway Rules to Interface Direction:  $prop_setting{'gatewaydir'}\n";
    print DUMPFILE "TCP Session Timeout (sec):                   $prop_setting{'tcptimeout'}\n";    
    print DUMPFILE "Accept Firewall-1 Control Connections:       $prop_setting{'fw1enable'} / $prop_setting{'fw1enable_p'}\n";
    print DUMPFILE "Accept UDP Replies:                          $prop_setting{'udpreply'} /  $prop_setting{'udpreply_p'}\n";
    print DUMPFILE "UDP Reply Timeout (sec):                     $prop_setting{'udptimeout'}\n";    
    print DUMPFILE "Accept Outgoing Packets:                     $prop_setting{'outgoing'} /  $prop_setting{'outgoing_p'}\n";
    print DUMPFILE "Enable Decryption on Accept:                 $prop_setting{'acceptdecrypt'}\n";
    print DUMPFILE "Use FASTPATH:                                $prop_setting{'enable_fastpath'}\n";
    print DUMPFILE "Accept RIP:                                  $prop_setting{'rip'} /  $prop_setting{'rip_p'}\n";
    print DUMPFILE "Accept Domain Name Queries (UDP):            $prop_setting{'domain_udp'} /  $prop_setting{'domain_udp_p'}\n";
    print DUMPFILE "Accept Domain Name Download (TCP):           $prop_setting{'domain_tcp'} /  $prop_setting{'domain_tcp_p'}\n";
    print DUMPFILE "Accept ICMP:                                 $prop_setting{'icmpenable'} / $prop_setting{'icmpenable_p'}\n";
    print DUMPFILE " \n";
    print DUMPFILE "Services\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "Enable FTP PASV Connections:                 $prop_setting{'ftppasv'} / $prop_setting{'ftppasv_p'}\n";
    print DUMPFILE "Enable RSH/REXEC Reverse stderr Connections: $prop_setting{'rshstderr'} / $prop_setting{'rshstderr_p'}\n";
    print DUMPFILE "Enable RPC Control:                          $prop_setting{'rpcenable'} / $prop_setting{'rpcenable_p'}\n";
    print DUMPFILE "Enable Response of FTP Data Connections:     $prop_setting{'ftpdata'} / $prop_setting{'ftpdata_p'}\n";
    print DUMPFILE "Enable Real Audio Reverse Connections:       $prop_setting{'raudioenable'} / $prop_setting{'raudioenable_p'}\n";
    print DUMPFILE "Enable VDOLive Reverse Connections:          $prop_setting{'vdolivenable'} / $prop_setting{'vdolivenable_p'}\n";
    print DUMPFILE "Enable CoolTalk Data Connections (UDP):      $prop_setting{'cooltalkenable'}\n";
    print DUMPFILE "Enable H.323 Control and Data Connections:   $prop_setting{'iphoneenable'}\n";
    print DUMPFILE " \n";
    print DUMPFILE "Log and Alert\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "Excessive Log Grace Period (sec):            $prop_setting{'loggrace'}\n";
    print DUMPFILE "PopUp Alert Command:                         $prop_setting{'alertcmd'}\n";
    print DUMPFILE "Mail Alert Command:                          $prop_setting{'mailcmd'}\n";
    print DUMPFILE "SNMP Trap Alert Command:                     $prop_setting{'snmptrapcmd'}\n";
    print DUMPFILE "User Defined Alert Command:                  $prop_setting{'useralertcmd'}\n";
    print DUMPFILE "Anti Spoof Alert Command:                    $prop_setting{'spoofalertcmd'}\n";
    print DUMPFILE "User Authentication Alert Command:           $prop_setting{'userauthalertcmd'}\n";    
    print DUMPFILE "Log Established TCP Packets:                 $prop_setting{'log_established_tcp'}\n";
    print DUMPFILE "Enable Active Connections:                   $prop_setting{'liveconns'}\n";
    print DUMPFILE " \n";
    print DUMPFILE "Resolving\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "Lookup Priorities:                        1. $prop_setting{'resolver_1'}\n";
    print DUMPFILE "                                          2. $prop_setting{'resolver_2'}\n";
    print DUMPFILE "                                          3. $prop_setting{'resolver_3'}\n";
    print DUMPFILE "                                          4. $prop_setting{'resolver_4'}\n";
    print DUMPFILE "BIND Timeout (sec):                          $prop_setting{'timeout'}\n";
    print DUMPFILE "BIND Retries:                                $prop_setting{'retries'}\n";
    print DUMPFILE "Log Viewer Resolver Properties:              $prop_setting{'pagetimeout'}\n";
    print DUMPFILE " \n";
    print DUMPFILE "Security Servers\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "Telnet Welcome Message File:                 $prop_setting{'telnet_msg'}\n";
    print DUMPFILE "Rlogin Welcome Message File:                 $prop_setting{'rlogin_msg'}\n";
    print DUMPFILE "FTP Welcome Message File:                    $prop_setting{'ftp_msg'}\n";
    print DUMPFILE "Client Authentication Welcome Message File:  $prop_setting{'clnt_auth_msg'}\n";
    print DUMPFILE "HTTP Next Proxy:                             $prop_setting{'http_next_proxy_host'} : $prop_setting{'http_next_proxy_port'}\n";
    print DUMPFILE " \n";
    print DUMPFILE "Authentication\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "User Authentication Session Timeout (min):   $prop_setting{'au_timeout'}\n";
    print DUMPFILE "AXENT Pathways Defender Server Setup:    IP: $prop_setting{'snk_server_ip'}\n";
    print DUMPFILE "                                   Agent ID: $prop_setting{'snk_agent_id'}\n";
    print DUMPFILE "                                  Agent Key: $prop_setting{'snk_agent_key'}\n";
    print DUMPFILE " \n";
    print DUMPFILE "SYNDefender\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "Method:                                      $SynDefender[$prop_setting{'fwsynatk_method'}]\n";
    print DUMPFILE "Timeout:                                     $prop_setting{'fwsynatk_timeout'}\n";
    print DUMPFILE "Maximum Sessions:                            $prop_setting{'fwsynatk_max'}\n";
    print DUMPFILE "Display Warning Messages:                    $prop_setting{'fwsynatk_warning'}\n";
    print DUMPFILE "</table>\n";
    print DUMPFILE " \n";
    print DUMPFILE "Miscellaneous\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "Load Agents Port:                            $prop_setting{'load_service_port'}\n";
    print DUMPFILE "Load Measurement Interval:                   $prop_setting{'lbalanced_period_wakeup_sec'}\n";
    print DUMPFILE " \n";
    print DUMPFILE "Access Lists\n";
    print DUMPFILE "---------------------------------------------------------\n";
    print DUMPFILE "Accept Established TCP Connections:          $prop_setting{'established'} / $prop_setting{'established_p'}\n";
    print DUMPFILE "Accept RIP:                                  $prop_setting{'rip'} / $prop_setting{'rip_p'}\n";
    print DUMPFILE "Accept Domain Name Queries (UDP):            $prop_setting{'domain_udp'} / $prop_setting{'domain_udp_p'}\n";
    print DUMPFILE "Accept Domain Name Download (TCP):           $prop_setting{'domain_tcp'} / $prop_setting{'domain_tcp_p'}\n";
    print DUMPFILE "Accept ICMP:                                 $prop_setting{'icmpenable'} / $prop_setting{'icmpenable_p'}\n";
    close (DUMPFILE);
}


##########################################################################
# print implicit rules in Text/TSV
#	file handle
#	location id ( first / "before last" / last
#	format string for output
sub subImplicit__Dump {
    my ($FILE)         = $_[0];
    my ($locstr)       = $_[1];
    my ($formstring)   = $_[2];
    
    if ( ("$prop_setting{'fw1enable'}" eq 'true') &&
         (lc("$prop_setting{'fw1enable_p'}") eq "$locstr") ) {
	printf DUMPFILE ("$formstring",'FW1 Host','FW1 Host','FW1','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FW1 Host','FW1 Host','FW1_log','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'gui_clients ','FW1 Management','FW1_mgmt','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FloodGate-1 Host','FW1 Management','FW1_ela','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'Any ','FW1 Host','FW1_topo','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'Any ','FW1 Host','FW1_key','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'Any ','FW1 Host','IKE','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FW1 Host','Any','IKE','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'Any ','Any','RDP','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FW1 Host','CVP-Servers','FW1_cvp','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FW1 Host','UFP-Servers','FW1_ufp','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FW1 Host','Radius-Servers','RADIUS','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FW1 Host','Tacacs-Servers','TACACS','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FW1 Host','Ldap-Servers','ldap','FW1 Control Connections');
	printf DUMPFILE ("$formstring",'FW1 Host','Logical-Servers','load_agent','FW1 Control Connections');
    }
# outgoing
    if ( ("$prop_setting{'outgoing'}" eq 'true') &&
         (lc("$prop_setting{'outgoing_p'}") eq "$locstr") ) {
	printf DUMPFILE ("$formstring",'FW1 Module','Any','Any','Outgoing');
    }
# RIP
    if ( ("$prop_setting{'rip'}" eq 'true') &&
         (lc("$prop_setting{'rip_p'}") eq "$locstr") ) {
	printf DUMPFILE ("$formstring",'Any ','Any','RIP','RIP');
    }

# ICMP
    if ( ("$prop_setting{'icmpenable'}" eq 'true') &&
         (lc("$prop_setting{'icmpenable_p'}") eq "$locstr") ) {
	printf DUMPFILE ("$formstring",'Any ','Any','icmp','ICMP');
    }
# domain TCP
    if ( ("$prop_setting{'domain_tcp'}" eq 'true') &&
         (lc("$prop_setting{'domain_tcp_p'}") eq "$locstr") ) {

	printf DUMPFILE ("$formstring",'Any ','Any','domain_tcp','Domain-TCP');
    }
# domain UDP
    if ( ("$prop_setting{'domain_udp'}" eq 'true') &&
         (lc("$prop_setting{'domain_udp_p'}") eq "$locstr") ) {
	printf DUMPFILE ("$formstring",'Any ','Any','domain_udp','Domain-UDP');
    }
}

##########################################################################
# dump access rules into a TXT file
#	filename for the dump

sub DumpRulesTXT{
    my ($filename) 	= $_[0];
    my ($i);
    my ($txt);

    open (DUMPFILE,">$filename") ;
    #--- first the explicit ones ---
    for ( $i = 0; $i<=$access_number; $i++ ) {
	    if ( $i == 0 ){
	    	subImplicit__Dump (DUMPFILE, 'first', "\nImplicit Rule\n\tFrom:       %s\n\tTo:         %s\n\tService:    %s\n\tAction:     accept	    \n\tTrack:      (none)\n\tTime:       Any\n\tInstall on: Gateways\n\tComment:    Enable %s\n");
            } elsif ( $i == $access_number ){
	    	subImplicit__Dump (DUMPFILE, 'before last', "\nImplicit Rule\n\tFrom:       %s\n\tTo:         %s\n\tService:    %s\n\tAction:     accept	    \n\tTrack:      (none)\n\tTime:       Any\n\tInstall on: Gateways\n\tComment:    Enable %s\n");
            }
	    print DUMPFILE "\nRule ";
	    print DUMPFILE $i+1;
	    if ( $access_disabled[$i] ) { print DUMPFILE " (disabled) "; }
	    print DUMPFILE "\n";

	    print DUMPFILE "\tFrom:       ";
	    if ( $access_from_negated[$i] ) { print DUMPFILE "(negated) "; }
	    $txt = "$access_from[$i]"; $txt =~ s// /g;
	    print DUMPFILE "$txt\n";
	    
	    print DUMPFILE "\tTo:         ";
	    if ( $access_to_negated[$i] ) { print DUMPFILE "(negated) "; }
	    $txt = "$access_to[$i]"; $txt =~ s// /g;
	    print DUMPFILE "$txt\n";
	    
	    print DUMPFILE "\tService:    ";
	    if ( $access_services_negated[$i] ) { print DUMPFILE "(negated) "; }
	    $txt = "$access_services[$i]"; $txt =~ s// /g;
	    print DUMPFILE "$txt\n";

	    print DUMPFILE "\tAction:     $access_action[$i]\n";
	    print DUMPFILE "\tTrack:      $access_track[$i]\n";
	    print DUMPFILE "\tTime:       $access_time[$i]\n";
	    &PrintNonempty(DUMPFILE,"\tInstall on: %s\n",$access_install_on[$i]);
	    &PrintNonempty(DUMPFILE,"\tComment:    %s\n",$access_comment[$i]);
    }
    #--- now the implicit ones ---
    subImplicit__Dump (DUMPFILE, 'last', "\nImplicit Rule\n\tFrom:       %s\n\tTo:         %s\n\tService:    %s\n\tAction:     accept	    \n\tTrack:      (none)\n\tTime:       Any\n\tInstall on: Gateways\n\tComment:    Enable %s\n");
    
    close (DUMPFILE);
}


##########################################################################
# dump access rules into a TSV file
#	filename for the dump


sub DumpRulesTSV{
    my ($filename) 	= $_[0];
    my ($txt);

    open (DUMPFILE,">$filename") ;
    print DUMPFILE "number\tdisabled\tfrom neg.\tfrom\tto neg.\tto\tsvc.neg.\tservice\taction\ttrack\ttime\tinst.on\tcomment\n";
    print DUMPFILE "-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\n";
    #--- first the explicit ones ---
    for ( $i = 0; $i<=$access_number; $i++ ) {
	    if ( $i == 0 ){
	    	subImplicit__Dump (DUMPFILE, 'first', "(I)\t\t%s\t%s\t%s\taccept\t(none)\tAny\tGateways\tImplicit Rule: Enable %s\n");
            } elsif ( $i == $access_number ){
	    	subImplicit__Dump (DUMPFILE, '"before last"', "(I)\t\t%s\t%s\t%s\taccept\t(none)\tAny\tGateways\tImplicit Rule: Enable %s\n");
            }
	    print DUMPFILE $i+1;
	    print DUMPFILE "\t";
	    if ( $access_disabled[$i] ) { print DUMPFILE "(disabled)"; }
	    print DUMPFILE "\t";
	    
	    if ( $access_from_negated[$i] ) { print DUMPFILE "(negated) "; }
	    $txt = "$access_from[$i]"; $txt =~ s// /g;
	    print DUMPFILE "$txt";
	    
	    if ( $access_to_negated[$i] ) { print DUMPFILE "(negated) "; }
	    $txt = "$access_to[$i]"; $txt =~ s// /g;
	    print DUMPFILE "$txt";
	    
	    if ( $access_services_negated[$i] ) { print DUMPFILE "(negated) "; }
	    $txt = "$access_services[$i]"; $txt =~ s// /g;
	    print DUMPFILE "$txt";
	    
	    print DUMPFILE "$access_action[$i]\t$access_track[$i]\t$access_time[$i]\t$access_install_on[$i]\t$access_comment[$i]\n";
    }
    #--- now the implicit ones ---
    subImplicit__Dump (DUMPFILE, 'last', "(I)\t\t%s\t%s\t%s\taccept\t(none)\tAny\tGateways\tImplicit Rule: Enable %s\n");
    close (DUMPFILE);
}


##########################################################################
# dump NAT rules into a TSV file
#	filename for the dump
sub DumpNatTSV{
    my ($filename) 	= $_[0];
    my ($i);

    open (DUMPFILE,">$filename") ;
    print DUMPFILE "number\timplicit\torig.from\torig.to\torig.service\ttrans.from\tmethod\ttrans.to\tmethod\ttrans.service\tinstall-on\tcomment\n";
    print DUMPFILE "-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\n";
    #--- first the explicit ones ---
    for ( $i = 0; $i<$nat_number; $i++ ) {
	    print DUMPFILE $i+1;
	    print DUMPFILE "\tno\t$nat_orig_from[$i]\t$nat_orig_to[$i]\t$nat_orig_svc[$i]\t";
	    print DUMPFILE "$nat_transl_from[$i]\t$NATtranslation{$nat_transl_from_methd[$i]}\t";
	    print DUMPFILE "$nat_transl_to[$i]\t$NATtranslation{$nat_transl_to_methd[$i]}\t";
	    print DUMPFILE "$nat_transl_svc[$i]\t$NATtranslation{$nat_transl_svc_methd[$i]}\t";
	    print DUMPFILE "$nat_install_on[$i]\t";
	    print DUMPFILE "$nat_comment[$i]\n";
    }
    #--- now the implicit ones ---
    if ( $FLAG_implicitrules ) {
	foreach $name (@obj_name){
	    if ( ( $obj_used{$name} || $FLAG_allobjs ) &&
	         ( "$obj_NATadr{$name}" ne '' ) ) {
		#--- forward rule
	        $i++;
		print DUMPFILE "$i\tyes\t$name\tAny\tAny\t$obj_NATadr{$name}\t";
		print DUMPFILE "$NATtranslation{$obj_NATtype{$name}}\t";
		print DUMPFILE "Original\t\tOriginal\t\t\n";
		#--- backward rule
	        $i++;
		print DUMPFILE "$i\tyes\tAny\t$obj_NATadr{$name}\tAny\tOriginal\t\t$name\t";
		print DUMPFILE "$NATtranslation{$obj_NATtype{$name}}\t";
		print DUMPFILE "Original\t\t\t\n";
	    }	
	}
    }
    close (DUMPFILE);
}


##########################################################################
# dump NAT rules into a TXT file
#	filename for the dump
#	switch:  1 = if with implicit rules,   0 = used only
#	switch:  1 = if all objects,   0 = used only

sub DumpNatTXT{
    my ($filename) 	= $_[0];
    my ($with_implicit) = $_[1];
    my ($if_used)  	= $_[2];
    my ($i);

    open (DUMPFILE,">$filename") ;
    #--- first the explicit ones ---
    for ( $i = 0; $i<$nat_number; $i++ ) {
	    print DUMPFILE "\nRule ";
	    print DUMPFILE $i+1;
	    print DUMPFILE " (explicit)\n";
	    print DUMPFILE "\tInstalled On: $nat_install_on[$i]\n";
	    print DUMPFILE "\tORIGINAL\n";
	    print DUMPFILE "\t\tFrom:    $nat_orig_from[$i]\n";
	    print DUMPFILE "\t\tTo:      $nat_orig_to[$i]\n";
	    print DUMPFILE "\t\tService: $nat_orig_svc[$i]\n";
	    print DUMPFILE "\tTRANSLATED\n";
	    print DUMPFILE "\t\tFrom:    $nat_transl_from[$i]";
	    print DUMPFILE " ($NATtranslation{$nat_transl_from_methd[$i]})\n";
	    print DUMPFILE "\t\tTo:      $nat_transl_to[$i]";
	    print DUMPFILE " ($NATtranslation{$nat_transl_to_methd[$i]})\n";
	    print DUMPFILE "\t\tService: $nat_transl_svc[$i]";
	    print DUMPFILE " ($NATtranslation{$nat_transl_svc_methd[$i]})\n";
	    &PrintNonempty(DUMPFILE,"\tComment: %s\n",$nat_comment[$i]);
    }
    #--- now the implicit ones ---
    if ( $FLAG_implicitrules ) {
	foreach $name (@obj_name){
	    if ( ( $obj_used{$name} || $FLAG_allobjs ) &&
	         ( "$obj_NATadr{$name}" ne '' ) ) {
		#--- forward rule
	        $i++;
		print DUMPFILE "\nRule $i (implicit)\n";
		print DUMPFILE "\tORIGINAL\n";
		print DUMPFILE "\t\tFrom:    $name\n";
		print DUMPFILE "\t\tTo:      Any\n";
		print DUMPFILE "\t\tService: Any\n";
		print DUMPFILE "\tTRANSLATED\n";
		print DUMPFILE "\t\tFrom:    $obj_NATadr{$name}";
		print DUMPFILE " ($NATtranslation{$obj_NATtype{$name}})\n";
		print DUMPFILE "\t\tTo:      Original\n";
		print DUMPFILE "\t\tService: Original\n";
		#--- backward rule
	        $i++;
		print DUMPFILE "\nRule $i (implicit)\n";
		print DUMPFILE "\tORIGINAL\n";
		print DUMPFILE "\t\tFrom:    Any\n";
		print DUMPFILE "\t\tTo:      $obj_NATadr{$name}\n";
		print DUMPFILE "\t\tService: Any\n";
		print DUMPFILE "\tTRANSLATED\n";
		print DUMPFILE "\t\tFrom:    Original\n";
		print DUMPFILE "\t\tTo:      $name";
		print DUMPFILE " ($NATtranslation{$obj_NATtype{$name}})\n";
		print DUMPFILE "\t\tService: Original\n";
	    }	
	}
    }
    close (DUMPFILE);
}


##########################################################################
# dump all objects into a TSV file
#	filename for the dump

sub DumpObjectsTSV{
    my ($filename) = $_[0];
    my ($name)     = '';

    open (DUMPFILE,">$filename") ;
    print DUMPFILE "name\ttype\tlocation\tis_fw1\tipaddr\tnetmask\tNATadr\tNATtype\tmembers\tcomment\tcolour\n";
    print DUMPFILE "-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\n";
    foreach $name (@obj_name){
	if ( $obj_used{$name} || $FLAG_allobjs ) {
	    print DUMPFILE "$name\t$obj_type{$name}\t$obj_location{$name}\t$obj_is_fw1{$name}\t$obj_ipaddr{$name}\t$obj_netmask{$name}\t$obj_NATadr{$name}\t$obj_NATtype{$name}\t$obj_members{$name}\t$obj_comment{$name}\t$obj_colour{$name}\n";
	}
    }
    close (DUMPFILE);
}


##########################################################################
# dump all objects into a nice text file
#	filename for the dump

sub DumpObjectsTXT{
    my ($filename) = $_[0];
    my ($name)     = '';

    open (DUMPFILE,">$filename") ;
    foreach $name (@obj_name){
	if ( $obj_used{$name} || $FLAG_allobjs ) {
	    print DUMPFILE "\n$name\n";
	    print DUMPFILE "\tType: $obj_type{$name}\n";
	    if ( $obj_location{$name} ) {
		print DUMPFILE "\tLocation: external\n";
	    } else {
		print DUMPFILE "\tLocation: internal\n";
	    }
	    if ( $obj_is_fw1{$name} ) {
		print DUMPFILE "\tFirewall-1 is installed\n";
	    }
	    &PrintNonempty(DUMPFILE,"\tIP-Address: %s\n",$obj_ipaddr{$name});
	    &PrintNonempty(DUMPFILE,"\tNetmask: %s\n",$obj_netmask{$name});
	    &PrintNonempty(DUMPFILE,"\tNAT-Address: %s ($NATtranslation{$obj_NATtype{$name}})\n",$obj_NATadr{$name});
	    &PrintNonempty(DUMPFILE,"\tMembers: %s\n",$obj_members{$name});
	    &PrintNonempty(DUMPFILE,"\tComment: %s\n",$obj_comment{$name});
	}
    }
    close (DUMPFILE);
}


##########################################################################
# dump all services into a TSV file
#	filename for the dump

sub DumpServicesTSV{
    my ($filename) = $_[0];
    my ($name)     = '';

    open (DUMPFILE,">$filename") ;
    print DUMPFILE "name\ttype\tdst_port\tsrc_port_low\tsrc_port_high\tmatch\tprolog\tmembers\tcomment\tcolour\n";
    print DUMPFILE "-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\n";
    foreach $name (@svc_name){
	if ( $svc_used{$name} || $FLAG_allservices ) {
	    print DUMPFILE "$name\t$svc_type{$name}\t$svc_dst_port{$name}\t$svc_src_port_low{$name}\t$svc_src_port_high{$name}\t$svc_match{$name}\t$svc_prolog{$name}\t$svc_members{$name}\t$svc_comment{$name}\t$svc_colour{$name}\n";
	}
    }
    close (DUMPFILE);
}

##########################################################################
# dump all services into a nice text file
#	filename for the dump

sub DumpServicesTXT{
    my ($filename) = $_[0];
    my ($name)     = '';

    open (DUMPFILE,">$filename") ;
    foreach $name (@svc_name){
	if ( $svc_used{$name} || $FLAG_allservices ) {
	    print DUMPFILE "\n$name\n";
	    print DUMPFILE "\tType: $svc_type{$name}\n";
	    &PrintNonempty(DUMPFILE,"\tDestination Port: %s\n",$svc_dst_port{$name});
	    &PrintNonempty(DUMPFILE,"\tSource Port (low): %s\n",$svc_src_port_low{$name});
	    &PrintNonempty(DUMPFILE,"\tSource Port (high): %s\n",$svc_src_port_high{$name});
	    &PrintNonempty(DUMPFILE,"\tpre-Match: %s\n",$svc_match{$name});
	    &PrintNonempty(DUMPFILE,"\tProlog: %s\n",$svc_prolog{$name});
	    &PrintNonempty(DUMPFILE,"\tMembers: %s\n",$svc_members{$name});
	    &PrintNonempty(DUMPFILE,"\tComment: %s\n",$svc_comment{$name});
	}
    }
    close (DUMPFILE);
}

##########################################################################
# dump all users into a TSV file
#	filename for the dump

sub DumpUsersTSV{
    my ($filename) = $_[0];
    my ($name)     = '';

    open (DUMPFILE,">$filename") ;
    print DUMPFILE "name\ttype\tfrom\tto\tauth\tday\texpires\tmembers\tcomment\tcolour\n";
    print DUMPFILE "-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\n";
    foreach $name (@user_name){
	next if($name =~ /^ *$/);
	next if($name =~ /All Users/);
	next if($name =~ /ALL_GROUPS/);
	next if($name =~ /ALL_KEYH/);
	next if($name =~ /ALL_TEMPLATES/);
	next if($user_type{$name} =~ /keyh/);
	next if($user_type{$name} =~ /template/);
	if ( $user_used{$name} || $FLAG_alluser ) {
	   print DUMPFILE "$name\t$user_type{$name}\t$user_from{$name}\t$user_to{$name}\t$user_auth{$name}\t$user_day{$name}\t$user_expires{$name}\t$user_members{$name}\t$user_comment{$name}";
	}
    }
    close (DUMPFILE);
}


##########################################################################
# dump all users into a nice text file
#	filename for the dump

sub DumpUsersTXT{
    my ($filename) = $_[0];
    my ($name)     = '';

    open (DUMPFILE,">$filename") ;
    foreach $name (@user_name){
	next if($name =~ /^ *$/);
	next if($name =~ /All Users/);
	next if($name =~ /ALL_GROUPS/);
	next if($name =~ /ALL_KEYH/);
	next if($name =~ /ALL_TEMPLATES/);
	next if($user_type{$name} =~ /keyh/);
	next if($user_type{$name} =~ /template/);
	if ( $user_used{$name} || $FLAG_alluser ) {
	   print DUMPFILE "$name\n";
	   &PrintNonempty(DUMPFILE,"\tType:    %s\n",$user_type{$name});
	   &PrintNonempty(DUMPFILE,"\tFrom:    %s\n",$user_from{$name});
	   &PrintNonempty(DUMPFILE,"\tTo:      %s\n",$user_to{$name});
	   &PrintNonempty(DUMPFILE,"\tAuth:    %s\n",$user_auth{$name});
	   &PrintNonempty(DUMPFILE,"\tDay:     %s\n",$user_day{$name});
	   &PrintNonempty(DUMPFILE,"\tExpires: %s\n",$user_expires{$name});
	   &PrintNonempty(DUMPFILE,"\tMembers: %s\n",$user_members{$name});
	   &PrintNonempty(DUMPFILE,"\tComment: %s\n",$user_comment{$name});
	}
    }
    close (DUMPFILE);
}


##########################################################################
# dump all unused objects and services into a nice text file
#	filename for the dump

sub DumpUnusedObjects{
    my ($filename) = $_[0];
    my ($name)     = '';

    open (DUMPFILE,">$filename") ;
    print DUMPFILE "\n----------------------\n";
    print DUMPFILE "Unused Network Objects\n";
    print DUMPFILE "----------------------\n";
    foreach $name (@obj_name){
	if ( ! $obj_used{$name} ) {
	    print DUMPFILE "\n$name  \($obj_ipaddr{$name} / $obj_netmask{$name}\) $obj_members{$name}\n";
	}
    }
    print DUMPFILE "\n\n----------------------------------\n";
    print DUMPFILE "Unused Services (maybe predefined)\n";
    print DUMPFILE "----------------------------------\n";
    foreach $name (@svc_name){
	if ( ! $svc_used{$name} ) {
	    print DUMPFILE "\n$name  \($svc_type{$name}/$svc_dst_port{$name}\) $svc_members{$name} : $svc_comment{$name}\n";
	}
    }
    print DUMPFILE "\n\n----------------------------------\n";
    print DUMPFILE "Unused Users\n";
    print DUMPFILE "----------------------------------\n";
    close (DUMPFILE);
    foreach $name (@user_name){
	next if($name =~ /^ *$/);
	next if($name =~ /All Users/);
	next if($name =~ /ALL_GROUPS/);
	next if($name =~ /ALL_KEYH/);
	next if($name =~ /ALL_TEMPLATES/);
	next if($user_type{$name} =~ /keyh/);
	next if($user_type{$name} =~ /template/);
	if ( ! $user_used{$name} ) {
	   print DUMPFILE "$name ($user_type{$name})  $user_comment{$name}\n";
	}
    }
}


##########################################################################
##########################################################################
###   Print nice config files
##########################################################################
##########################################################################

##########################################################################
# give (translated) colour for HTML
#	color to set

sub HTMLcolour {
    my ($givencolour)   = $_[0];

    if ( ("$givencolour" eq '') ||  $FLAG_withcolors ) {
	if ( $colors{"$givencolour"} ne '' ) {
	    return("color=$colors{\"$givencolour\"}");
    	} else {
	    return("color=$givencolour");
	}
    } else {
    	return ('');
    }
}


##########################################################################
# print object or dash
#	file handle
#	object string (maybe null, in which case a dash is printed

sub PrintOrDash {
    my ($FILE)         = $_[0];
    my ($object)       = $_[1];
    
    if ( "$object" eq '') {
	print $FILE '-';
    } else {
	print $FILE "$object";
    }
}

##########################################################################
# print object for HTML
#	file handle
#	switch:  1 = negated,   0 = normal
#	object string (maybe multiple instances)

sub subObject__Output_in_HTML {
    my ($FILE)         = $_[0];
    my ($if_negated)   = $_[1];
    my ($objects)      = $_[2];
    my (@obj_array);
    my ($iconname)     = '';
    my ($colr)         = '';
    my ($neg)          = '';

    @obj_array = split (//, $objects);
    foreach $name (@obj_array){
	$colr = &HTMLcolour($obj_colour{$name});
	if ( $if_negated ) { 
	    $iconname = lc("$IconPathName/negated$obj_type{$name}.gif");
	    $neg = '!&nbsp;';
	} else {
	    $iconname = lc("$IconPathName/$obj_type{$name}.gif");
	}
	if ( ( lc("$name") eq 'any' ) || ( lc("$name") eq 'original' ) ) { 
	    print $FILE "\<img src=$IconPathName/any.gif\>\&nbsp;$name<br>";
	} elsif ( lc("$name") eq 'gateways' ) { 
	    print $FILE "\<img src=$IconPathName/gateways.gif\>\&nbsp;\Gateways\<br>";
	} else {
	    $addressinfo = "";
            if ( $FLAG_withip  && ($obj_ipaddr{$name} ne "") ) {
		if ($obj_netmask{$name} ne "") { 
		   $addressinfo =  "<br> net  " .  $obj_ipaddr{$name} . " / " . $netmasktranslation{"$obj_netmask{$name}"};
		} else {
		   $addressinfo =  "<br> host  " .  $obj_ipaddr{$name}; 
		}
	    }
            print $FILE "\<img src=$iconname\>\&nbsp;$neg\<a href=#$name\>\<font $colr\>$name\</font\>\</A\> $addressinfo<BR>";
	}
    }
}

##########################################################################
# print service for HTML
#	file handle
#	switch:  1 = negated,   0 = normal
#	object string (maybe multiple instances)

sub subService__Output_in_HTML {
    my ($FILE)         = $_[0];
    my ($if_negated)   = $_[1];
    my ($services)     = $_[2];
    my (@serv_array);
    my ($iconname)     = '';
    my ($neg)          = '';
    my ($ressource)    = '';

    @serv_array = split (//, $services);
    foreach $name (@serv_array){
	$colr = &HTMLcolour($svc_colour{$name});
    	$ressource = '';
	if ( $name =~ m/-\>/ ) {
	   ($name,$ressource) = split(/-\>/, $name, 2);
	}
	if ( $if_negated ) { 
	    $iconname = lc("$IconPathName/negated$svc_type{$name}.gif");
	    $neg = '!&nbsp;';
	} else {
	    $iconname = lc("$IconPathName/$svc_type{$name}.gif");
	}
	if ( ( lc("$name") eq 'any' ) || ( lc("$name") eq 'original' ) ) { 
	    print $FILE "\<img src=$IconPathName/any.gif\>\&nbsp;$name<br>";
	} else {
	    $aname = $name; $aname =~ s/>//g;
	    print $FILE "\<img src=$iconname\>\&nbsp;$neg\<a href=#$aname\>\<font $colr\>$name";
	    if ( "$ressource" ne '' ) {
		print $FILE "-&gt;\<a href=#$ressource\>$ressource\</a\>";
	    }
	    print $FILE "\</font\>\</a\><br>";
	}
    }
}

##########################################################################
# print userobject for HTML
#	file handle
#	switch:  1 = negated,   0 = normal
#	object string (maybe multiple instances)

sub subUserObject__Output_in_HTML {
    my ($FILE)                = $_[0];
    my ($if_negated)   = $_[1];
    my ($objects)      = $_[2];
    my (@obj_array);
    my ($iconname)     = '';
    my ($colr)         = '';
    my ($neg)          = '';
    my ($obj)          = '';
    my ($user)          = '';

    @obj_array = split (/ /, $objects);
    foreach $obj (@obj_array){
	($user,$name)=split (/@/, $obj);
	if ( ! $FLAG_nousers ) {
	   $colr = &HTMLcolour($user_colour{$user});
	   if ( $if_negated ) { 
	      $iconname = lc("$IconPathName/negated$user_type{$user}.gif");
	      $neg = '!&nbsp;';
	   } else {
	      $iconname = lc("$IconPathName/$user_type{$user}.gif");
	   }
	   if ( lc("$user") eq 'any' ) { 
	      print $FILE "\<img src=$iconname\>\&nbsp;$neg$user@";
	   } else {
	      print $FILE "\<img src=$iconname\>\&nbsp;$neg\<a href=#$user\>\<font $colr\>$user\</font\>\</a\>@";
	   }
	   if ( lc("$name") eq 'any' ) { 
	      print $FILE "$name\<br>";
	   } else {
	      print $FILE "\<a href=#$name\>\<font $colr\>$name\</font\>\</a\><br>";
	   }
	} else {
	   if ( $if_negated ) {
	      $iconname = lc("$IconPathName/negatedusergroup.gif");
	      $neg = '!&nbsp;';
	   } else {
	      $iconname = lc("$IconPathName/usergroup.gif");
	   }
	   print $FILE "\<img src=$iconname\>\&nbsp;$neg$user@";
	   print $FILE "$name\<br>";
	}
    }
}

##########################################################################
# print user for HTML
#	file handle
#	object string (maybe multiple instances)

sub subUser__Output_in_HTML {
    my ($FILE)         = $_[0];
    my ($user)     = $_[1];
    my (@user_array);
    my ($iconname)     = '';
    @user_array = split (/ /, $user);
    foreach $name (@user_array){
	next if($name =~ /^ *$/);
	$colr = &HTMLcolour($user_colour{$name});
	$iconname = lc("$IconPathName/$user_type{$name}.gif");
	if ( lc("$name") eq 'any' ) { 
	   print $FILE "\<img src=$IconPathName/any.gif\>\&nbsp;Any<br>";
	} else {
	   print $FILE "\<img src=$iconname\>\&nbsp;$neg\<a href=#$name\>\<font $colr\>$name\</font\>\</a\><br>";
	}
    }
}

##########################################################################
# print implicit rules in HTML
#	file handle
#	location id ( first / "before last" / last
sub subImplicit__Output_in_HTML {
    my ($FILE)         = $_[0];
    my ($locstr)       = $_[1];
    
    if ( ("$prop_setting{'fw1enable'}" eq 'true') &&
         (lc("$prop_setting{'fw1enable_p'}") eq "$locstr") ) {
	# FW1_mgmt
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/group.gif>\&nbsp;gui_clients</td><td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Management</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;FW1_mgmt</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# FW1_ela
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FloodGate-1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Management</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;FW1_ela</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# RDP
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/udp.gif>\&nbsp;RDP</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# FW1_cvp
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;CVP-Servers</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;FW1_cvp</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# FW1_ufp
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;UFP-Servers</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;FW1_ufp</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# RADIUS
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Radius-Servers</td><td bgcolor=#ffff00><img src=$IconPathName/udp.gif>\&nbsp;RADIUS</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# TACACS
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;TACACS-Servers</td><td bgcolor=#ffff00><img src=$IconPathName/udp.gif>\&nbsp;TACACS</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# ldap
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>';
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;LDAP-Servers</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;ldap</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# load_agent
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>';
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Logical-Servers</td><td bgcolor=#ffff00><img src=$IconPathName/other.gif>\&nbsp;load_agent</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# ike
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/udp.gif>\&nbsp;IKE</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# FW1_topo
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;FW1_topo</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# FW1_key
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;FW1_key</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# IKE
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/udp.gif>\&nbsp;IKE</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# FW1
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;FW1</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
	# FW1_log
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;FW1_log</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable FW1 </td><\/tr>\n";
    }
# outgoing
    if ( ("$prop_setting{'outgoing'}" eq 'true') &&
         (lc("$prop_setting{'outgoing_p'}") eq "$locstr") ) {
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/gateway.gif>\&nbsp;FW1 Host</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable Outgoing Connections </td><\/tr>\n";
    }
# RIP
    if ( ("$prop_setting{'rip'}" eq 'true') &&
         (lc("$prop_setting{'rip_p'}") eq "$locstr") ) {
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/udp.gif>\&nbsp;RIP</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable RIP </td><\/tr>\n";
    }

# ICMP
    if ( ("$prop_setting{'icmpenable'}" eq 'true') &&
         (lc("$prop_setting{'icmpenable_p'}") eq "$locstr") ) {
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/icmp.gif>\&nbsp;ICMP</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable ICMP </td><\/tr>\n";
    }
# domain TCP
    if ( ("$prop_setting{'domain_tcp'}" eq 'true') &&
         (lc("$prop_setting{'domain_tcp_p'}") eq "$locstr") ) {
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/tcp.gif>\&nbsp;domain-tcp</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule:  Enable Domain-TCP </td><\/tr>\n";
    }
# domain UDP
    if ( ("$prop_setting{'domain_udp'}" eq 'true') &&
         (lc("$prop_setting{'domain_udp_p'}") eq "$locstr") ) {
	print DUMPFILE '<tr valign=top><td bgcolor=#ffff00> &nbsp; </td>'; 
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/udp.gif>\&nbsp;domain-udp</td>";  # svc
	print DUMPFILE "<td bgcolor=#ffff00><img src=$IconPathName/accept.gif>\&nbsp;accept</td><td bgcolor=#ffff00> &nbsp; </td><td bgcolor=#ffff00><img src=$IconPathName/any.gif>\&nbsp;Any</td><td bgcolor=#ffff00><img src=$IconPathName/gateways.gif>\&nbsp;Gateways</td><td bgcolor=#ffff00>Implicit rule: Enable Domain-UDP </td><\/tr>\n";
    }
}


##########################################################################
# print configuration into HTML file
#	filename for the resulting file

sub Output_in_HTML {
    my ($filename)     = $_[0];
    my ($name)         = '';

    open (DUMPFILE,">$filename") ;
    print DUMPFILE '<html><HEAD>'; print DUMPFILE "\n";
    print DUMPFILE '<style type="text/css">'; print DUMPFILE "\n";
    print DUMPFILE '<!--'; print DUMPFILE "\n";
    print DUMPFILE '	BODY,TD,TH {'; print DUMPFILE "\n";
    print DUMPFILE '		font-family: Lucida,Verdana,Helvetica,Arial;'; print DUMPFILE "\n";
    print DUMPFILE '		font-size: 7 pt;'; print DUMPFILE "\n";
    print DUMPFILE '	}'; print DUMPFILE "\n";
    print DUMPFILE '-->'; print DUMPFILE "\n";
    print DUMPFILE '</style>'; print DUMPFILE "\n";
    print DUMPFILE '<title>'; print DUMPFILE "$HTMLtitle"; print DUMPFILE '</title>'; print DUMPFILE "\n";
    print DUMPFILE '</head>'; print DUMPFILE "\n";
    print DUMPFILE '<body bgcolor="#ffffff">'; print DUMPFILE "\n";
    print DUMPFILE '<h2>Firewall Policy: '; 
    print DUMPFILE "$FW1rules";
    print DUMPFILE '</h2>'; print DUMPFILE "\n";
    if ( ! $FLAG_norules ) {
      print DUMPFILE '<table border=1 cellpadding=5 bgcolor="Silver">'; print DUMPFILE "\n";
      print DUMPFILE '<tr valign=top><th bgcolor=#0000ff><FONT COLOR=white>RULE</FONT></th><th bgcolor=#0000ff><FONT COLOR=white>SOURCE</FONT></th><th bgcolor=#0000ff><FONT COLOR=white>DESTINATION</FONT></th><th bgcolor=#0000ff><FONT COLOR=white>SERVICES</FONT></th><th bgcolor=#0000ff><FONT COLOR=white>ACTION</FONT></th><th bgcolor=#0000ff><FONT COLOR=white>TRACK</FONT></th><th bgcolor=#0000ff><FONT COLOR=white>TIME</FONT></th><th bgcolor=#0000ff><FONT COLOR=white>INSTALL ON</FONT></th><th bgcolor=#0000ff><FONT COLOR=white>COMMENTS</FONT></th></tr>'; print DUMPFILE "\n";
      print DUMPFILE '<!---+++++++++++++++++++++++++++++++++++++++++++++--->'; print DUMPFILE "\n";
      for ( $i = 0; $i<=$access_number; $i++ ) {
	    if ( $i == 0 ){
	    	subImplicit__Output_in_HTML (DUMPFILE, 'first');
            } elsif ( $i == $access_number ){
	    	subImplicit__Output_in_HTML (DUMPFILE, 'before last');
            }
	    if ( $access_disabled[$i] ) { 
		print DUMPFILE '<tr valign=top><td bgcolor="#ff0000" ALIGN=center VALIGN=middle><FONT COLOR=white><B>';
		print DUMPFILE $i+1;
	    	print DUMPFILE '<BR><BR>DIS-<BR>ABLED</B></FONT></td>';
	    } else {
		print DUMPFILE '<tr valign=top><td bgcolor=#0000ff ALIGN=center VALIGN=middle><FONT COLOR=white><B>';
		print DUMPFILE $i+1;
		print DUMPFILE '</B></FONT></td>';
	    }
	    print DUMPFILE '<td>';
	    if($access_from[$i]=~/@/) {
		&subUserObject__Output_in_HTML (DUMPFILE, $access_from_negated[$i], $access_from[$i]);
	    } else {
		&subObject__Output_in_HTML (DUMPFILE, $access_from_negated[$i], $access_from[$i]);
	    }
	    
	    print DUMPFILE '<td>';
	    &subObject__Output_in_HTML (DUMPFILE, $access_to_negated[$i], $access_to[$i]);
	    print DUMPFILE '</td>';
	    
	    print DUMPFILE '<td>';	    
	    &subService__Output_in_HTML (DUMPFILE, $access_services_negated[$i], $access_services[$i]);
	    print DUMPFILE '</td>';
	    
	    print DUMPFILE '<td>';	    
	    print DUMPFILE "\<img src=$IconPathName/$access_action[$i].gif\>\&nbsp;$access_action[$i]";
	    print DUMPFILE '</td>';
	    
	    print DUMPFILE '<td>';
	    if ( "$access_track[$i]" ne '' ){
		$name = lc("$access_track[$i].gif");
		print DUMPFILE "\<img src=$IconPathName/$name\>\&nbsp;$access_track[$i]";
	    }
	    print DUMPFILE '&nbsp;</td>';
	    
	    print DUMPFILE '<td>';	    
	    &subObject__Output_in_HTML (DUMPFILE, 0, $access_time[$i]);
	    print DUMPFILE '&nbsp;</td>';
	    
	    print DUMPFILE '<td>';	    
	    &subObject__Output_in_HTML (DUMPFILE, 0, $access_install_on[$i]);
	    print DUMPFILE '&nbsp;</td>';
	    
	    print DUMPFILE '<td>';	    
	    print DUMPFILE "$access_comment[$i]";
	    print DUMPFILE '&nbsp;</td></tr>';	    
	    print DUMPFILE "\n";	    
      }
      subImplicit__Output_in_HTML (DUMPFILE, 'last');
      print DUMPFILE '<!--------------------------------------------------->'; print DUMPFILE "\n";
      print DUMPFILE '</table><p>'; print DUMPFILE "\n";
      print DUMPFILE ''; print DUMPFILE "\n";
      print DUMPFILE '<p>&nbsp;<p>'; print DUMPFILE "\n";
      print DUMPFILE '<h3>Address Translation Rules</h3>'; print DUMPFILE "\n";
      print DUMPFILE '<table border=1 cellpadding=5 bgcolor="Silver"><tr valign=top><th rowspan=2 bgcolor=#0000ff><FONT COLOR=white>RULE</FONT></th><th colspan=3 bgcolor=#0000ff><FONT COLOR=white>ORIGINAL PACKET</FONT></th><th colspan=3 bgcolor=#0000ff><FONT COLOR=white>TRANSLATED PACKET</FONT></th><th rowspan=2 bgcolor=#0000ff><FONT COLOR=white>INSTALL ON</FONT></th><th rowspan=2 bgcolor=#0000ff><FONT COLOR=white>COMMENT</FONT></th></tr>'; print DUMPFILE "\n";
      print DUMPFILE '<tr valign=top><th bgcolor=#0000ff><font color=white>SOURCE</font></th><th bgcolor=#0000ff><font color=white>DESTINATION</font></th><th bgcolor=#0000ff><font color=white>SERVICE</font></th><th bgcolor=#0000ff><font color=white>SOURCE</font></th><th bgcolor=#0000ff><font color=white>DESTINATION</font></th><th bgcolor=#0000ff><font color=white>SERVICE</font></th></tr>'; print DUMPFILE "\n";
      print DUMPFILE '<!---+++++++++++++++++++++++++++++++++++++++++++++--->'; print DUMPFILE "\n";
      #--- first the explicit ones ---
      for ( $i = 0; $i<$nat_number; $i++ ) {
	    print DUMPFILE '<tr valign=top><td bgcolor=#0000ff align=center valign=middle><font color=white><B>';
	    print DUMPFILE $i+1;
	    print DUMPFILE '</B></font></td><td>';
	    &subObject__Output_in_HTML (DUMPFILE, 0, $nat_orig_from[$i]);
	    print DUMPFILE '</td><td>';
	    &subObject__Output_in_HTML (DUMPFILE, 0, $nat_orig_to[$i]);
	    print DUMPFILE '</td><td>';
	    &subService__Output_in_HTML (DUMPFILE, 0, $nat_orig_svc[$i]);
	    print DUMPFILE '</td><td>';
	    if ( $nat_transl_from_methd[$i] ) { print DUMPFILE 'S: '; } else { print DUMPFILE 'H: '; }
	    &subObject__Output_in_HTML (DUMPFILE, 0, $nat_transl_from[$i]);
	    print DUMPFILE '</td><td>';
	    if ( $nat_transl_from_methd[$i] ) { print DUMPFILE 'S: '; } else { print DUMPFILE 'H: '; }
	    &subObject__Output_in_HTML (DUMPFILE, 0, $nat_transl_to[$i]);
	    print DUMPFILE '</td><td>';
	    if ( $nat_transl_from_methd[$i] ) { print DUMPFILE 'S: '; } else { print DUMPFILE 'H: '; }
	    &subService__Output_in_HTML (DUMPFILE, 0, $nat_transl_svc[$i]);
	    print DUMPFILE '</td><td>';
	    &subObject__Output_in_HTML (DUMPFILE, 0, $nat_install_on[$i]);
	    print DUMPFILE "\</td\>\<td\>$nat_comment[$i]\&nbsp;\</td\>\</tr\>\n";
      }
      #--- now the implicit ones ---
      if ( $FLAG_implicitrules ) {
	foreach $name (@obj_name){
	    if ( ( $obj_used{$name} || $FLAG_allobjs ) &&
	         ( "$obj_NATadr{$name}" ne '' ) ) {
		#--- forward rule
	        $i++;
		print DUMPFILE '<tr valign=top><td bgcolor=#ffff00 align=center valign=middle><B>';
		print DUMPFILE $i;
		print DUMPFILE '</B></td><td bgcolor=#ffff00>';
		&subObject__Output_in_HTML (DUMPFILE, 0, $name);
		print DUMPFILE '</td>';
		print DUMPFILE "\<td bgcolor=#ffff00\>\<img src=$IconPathName/any.gif\>\&nbsp;Any\</td\>";
		print DUMPFILE "\<td bgcolor=#ffff00\>\<img src=$IconPathName/any.gif\>\&nbsp;Any\</td\>";
		print DUMPFILE '<td bgcolor=#ffff00>';
		if ( $obj_NATtype{$name} ) { print DUMPFILE 'S: '; } else { print DUMPFILE 'H: '; }
		print DUMPFILE "$obj_NATadr{$name}";
		print DUMPFILE '</td>';
		print DUMPFILE "\<td bgcolor=#ffff00\>\<img src=$IconPathName/any.gif\>\&nbsp;Original\</td\>";
		print DUMPFILE "\<td bgcolor=#ffff00\>\<img src=$IconPathName/any.gif\>\&nbsp;Original\</td\>";
		print DUMPFILE '<td bgcolor=#ffff00>&nbsp;</td>';
		print DUMPFILE "\<td bgcolor=#ffff00\>\(implicit rule set in object definition\)\</td\>\</tr\>\n";
		#--- backward rule
	        $i++;
		print DUMPFILE '<tr valign=top><td bgcolor=#ffff00 align=center valign=middle><B>';
		print DUMPFILE $i;
		print DUMPFILE '</B></td>';
		print DUMPFILE "\<td bgcolor=#ffff00\>\<img src=$IconPathName/any.gif\>\&nbsp;Any\</td\>";
		print DUMPFILE '<td bgcolor=#ffff00>';
		&subObject__Output_in_HTML (DUMPFILE, 0, $name);
		print DUMPFILE '</td>';
		print DUMPFILE "\<td bgcolor=#ffff00\>\<img src=$IconPathName/any.gif\>\&nbsp;Any\</td\>";
		print DUMPFILE "\<td bgcolor=#ffff00\>\<img src=$IconPathName/any.gif\>\&nbsp;Original\</td\>";
		print DUMPFILE '<td bgcolor=#ffff00>';
		if ( $obj_NATtype{$name} ) { print DUMPFILE 'S: '; } else { print DUMPFILE 'H: '; }
		print DUMPFILE "$obj_NATadr{$name}";
		print DUMPFILE '</td>';
		print DUMPFILE "\<td bgcolor=#ffff00\>\<img src=$IconPathName/any.gif\>\&nbsp;Original\</td\>";
		print DUMPFILE "\<td bgcolor=#ffff00\>\&nbsp;\</td\>\<td bgcolor=#ffff00\>\(implicit rule set in object definition\)\</td\>\</tr\>\n";
	    }	
	}
      }
      print DUMPFILE '<!--------------------------------------------------->'; print DUMPFILE "\n";
      print DUMPFILE '</table>'; print DUMPFILE "\n";
      print DUMPFILE '<br>LEGEND:  Address Translation Methods: S: Static, H: Hide<p>'; print DUMPFILE "\n";
      print DUMPFILE '<p>&nbsp;<p>'; print DUMPFILE "\n";
    }
    if ( ! $FLAG_noobjs ) {
      print DUMPFILE ''; print DUMPFILE "\n";
      print DUMPFILE '<h2>FireWall-1 Object Definitions</h2>'; print DUMPFILE "\n";
      print DUMPFILE ''; print DUMPFILE "\n";
      print DUMPFILE '<h3>Network Objects</h3>'; print DUMPFILE "\n";
      print DUMPFILE '<table border=2 cellpadding=5 bgcolor="Silver"><tr valign=top><th bgcolor=#0000ff><font color=white><B>Name</B></font></th><th bgcolor=#0000ff><font color=white><B>Type</B></font></th><th bgcolor=#0000ff><font color=white><B>Location</B></font></th><th bgcolor=#0000ff><font color=white><B>FW-1</B></font></th><th bgcolor=#0000ff><font color=white><B>IP Address</B></font></th><th bgcolor=#0000ff><font color=white><B>Netmask</B></font></th><th bgcolor=#0000ff><font color=white><B>NAT Address</B></font></th><th bgcolor=#0000ff><font color=white><B>Members</B></font></th><th bgcolor=#0000ff><font color=white><B>Comment</B></font></th></tr>'; print DUMPFILE "\n";
      print DUMPFILE '<!---+++++++++++++++++++++++++++++++++++++++++++++--->'; print DUMPFILE "\n";
      foreach $name (@obj_name){
	if ( $obj_used{$name} || $FLAG_allobjs ) {
	    print DUMPFILE '<tr valign=top><td><a NAME="';
	    print DUMPFILE "$name";
	    print DUMPFILE '"></a>';
	    print DUMPFILE "$name\</td\>\<td\>";
	    print DUMPFILE "\<img src=$IconPathName/$obj_type{$name}.gif\>\&nbsp;$obj_type{$name}";
	    print DUMPFILE '</td><td>';
	    if ( $obj_location{$name} ) {
		print DUMPFILE 'External</td><td>';
	    } else {
		print DUMPFILE 'Internal</td><td>';
	    }
	    if ( $obj_is_fw1{$name} ) {
		print DUMPFILE 'Yes</td><td>';
	    } else {
		print DUMPFILE 'No</td><td>';
	    }
	    &PrintOrDash(DUMPFILE,$obj_ipaddr{$name});
	    print DUMPFILE '</td><td>';
	    &PrintOrDash(DUMPFILE,$obj_netmask{$name});
	    print DUMPFILE '</td><td>';
            if ( "$obj_NATadr{$name}" ne '' ) {
		if ( $obj_NATtype{$name} ) {
 			print DUMPFILE 'S:';
		} else {
			print DUMPFILE 'H:';
 		}
                print DUMPFILE '&nbsp;';
                print DUMPFILE "$obj_NATadr{$name}";
 	    } else {
                print DUMPFILE '-';
            }
 	    print DUMPFILE '</td><td>';
	    if ( "$obj_members{$name}" eq '' ) {
		print DUMPFILE '&nbsp;';
	    } else {
		&subObject__Output_in_HTML (DUMPFILE, 0, $obj_members{$name});
	    }
	    print DUMPFILE "\</td\>\<td\>$obj_comment{$name}&nbsp;\</td\></tr>\n";
	}
      }
      print DUMPFILE '<!--------------------------------------------------->'; print DUMPFILE "\n";
      print DUMPFILE '</table>'; print DUMPFILE "\n";
      print DUMPFILE '<br>LEGEND:  Address Translation Methods: S: Static, H: Hide<p>'; print DUMPFILE "\n";
    }
    if ( ! $FLAG_noservices ) {
      print DUMPFILE ''; print DUMPFILE "\n";
      print DUMPFILE '<p>'; print DUMPFILE "\n";
      print DUMPFILE '<h3>Service Objects</h3>'; print DUMPFILE "\n";
      print DUMPFILE '<table border=2 cellpadding=5 bgcolor="Silver"><tr valign=top><th bgcolor=#0000ff><font color=white><B>Name</B></font></th><th bgcolor=#0000ff><font color=white><B>Type</B></font></th><th bgcolor=#0000ff><font color=white><B>Port/<br>Program</B></font></th><th bgcolor=#0000ff><font color=white><B>S_Port from</B></font></th><th bgcolor=#0000ff><font color=white><B>S_Port to:</B></font></th><th bgcolor=#0000ff><font color=white><B>Match</B></font></th><th bgcolor=#0000ff><font color=white><B>Prolog</B></font></th><th bgcolor=#0000ff><font color=white><B>Members</B></font></th><th bgcolor=#0000ff><font color=white><B>Comment</B></font></th></tr>'; print DUMPFILE "\n";
      print DUMPFILE '<!---+++++++++++++++++++++++++++++++++++++++++++++--->'; print DUMPFILE "\n";
      foreach $name (@svc_name){
	if ( $svc_used{$name} || $FLAG_allservices ) {
	    print DUMPFILE '<tr valign=top><td><a NAME="';
	    print DUMPFILE "$name";
	    print DUMPFILE '"></a>';
	    print DUMPFILE "$name";
	    print DUMPFILE '</td><td>';
	    print DUMPFILE "\<img src=$IconPathName/$svc_type{$name}.gif\>\&nbsp;$svc_type{$name}";
	    print DUMPFILE '</td><td>';
	    &PrintOrDash(DUMPFILE,$svc_dst_port{$name});
	    print DUMPFILE '</td><td>';
	    &PrintOrDash(DUMPFILE,$svc_src_port_low{$name});
	    print DUMPFILE '</td><td>';
	    &PrintOrDash(DUMPFILE,$svc_src_port_high{$name});
	    print DUMPFILE '</td><td>';
	    &PrintOrDash(DUMPFILE,$svc_match{$name});
	    print DUMPFILE '</td><td>';
	    &PrintOrDash(DUMPFILE,$svc_prolog{$name});
	    print DUMPFILE '</td><td>';
	    if ( "$svc_members{$name}" eq '' ) {
		print DUMPFILE '&nbsp;';
	    } else {
		&subService__Output_in_HTML (DUMPFILE, 0, $svc_members{$name});
	    }
	    print DUMPFILE "\</td\>\<td\>$svc_comment{$name}&nbsp;\</td\></tr>\n";
	}
      }
      print DUMPFILE '<!--------------------------------------------------->'; print DUMPFILE "\n";
      print DUMPFILE '</table>'; print DUMPFILE "\n";
    }
    if ( ! $FLAG_nousers ) {
      print DUMPFILE ''; print DUMPFILE "\n";
      print DUMPFILE '<p>'; print DUMPFILE "\n";
      print DUMPFILE '<h3>User Objects</h3>'; print DUMPFILE "\n";
      print DUMPFILE '<table border=2 cellpadding=5 bgcolor="Silver"><tr valign=top><th bgcolor=#0000ff><font color=white><B>Name</B></font></th><th bgcolor=#0000ff><font color=white><B>Type</B></font></th><th bgcolor=#0000ff><font color=white><B>From</B></font></th><th bgcolor=#0000ff><font color=white><B>To</B></font></th><th bgcolor=#0000ff><font color=white><B>Auth</B></font></th><th bgcolor=#0000ff><font color=white><B>Day</B></font></th><th bgcolor=#0000ff><font color=white><B>Expires</B></font></th><th bgcolor=#0000ff><font color=white><B>Members</B></font></th><th bgcolor=#0000ff><font color=white><B>Comment</B></font></th></tr>'; print DUMPFILE "\n";
      print DUMPFILE '<!---+++++++++++++++++++++++++++++++++++++++++++++--->'; print DUMPFILE "\n";
      foreach $name (@user_name){
	next if($name =~ /^ *$/);
	next if($name =~ /All Users/);
	next if($name =~ /ALL_GROUPS/);
	next if($name =~ /ALL_KEYH/);
	next if($name =~ /ALL_TEMPLATES/);
	next if($user_type{$name} =~ /keyh/);
	next if($user_type{$name} =~ /template/);
	if ( $user_used{$name} || $FLAG_alluser ) {
	   print DUMPFILE '<tr valign=top><td><a NAME="';
	   print DUMPFILE "$name";
	   print DUMPFILE '"></a>';
	   print DUMPFILE "$name";
	   print DUMPFILE '</td><td>';
	   print DUMPFILE "\<img src=$IconPathName/$user_type{$name}.gif\>\&nbsp;$user_type{$name}";
	   print DUMPFILE '</td><td>';
	   &PrintOrDash(DUMPFILE,$user_from{$name});
	   print DUMPFILE '</td><td>';
	   &PrintOrDash(DUMPFILE,$user_to{$name});
	   print DUMPFILE '</td><td>';
	   &PrintOrDash(DUMPFILE,$user_auth{$name});
	   print DUMPFILE '</td><td>';
	   &PrintOrDash(DUMPFILE,$user_day{$name});
	   print DUMPFILE '</td><td>';
	   &PrintOrDash(DUMPFILE,$user_expires{$name});
	   print DUMPFILE '</td><td>';
	   if ( "$user_members{$name}" eq '' ) {
	       print DUMPFILE '&nbsp;';
	   } else {
	       &subUser__Output_in_HTML (DUMPFILE, $user_members{$name});
	   }
	   print DUMPFILE "\</td\>\<td\>$user_comment{$name}&nbsp;\</td\></tr>\n";
	}
      }
      print DUMPFILE '<!--------------------------------------------------->'; print DUMPFILE "\n";
      print DUMPFILE '</table>'; print DUMPFILE "\n";
      print DUMPFILE ''; print DUMPFILE "\n";
    }
    print DUMPFILE '<p>&nbsp;<p>'; print DUMPFILE "\n";
    print DUMPFILE '<h3>Property Settings</h3>'; print DUMPFILE "\n";
    print DUMPFILE '<!---+++++++++++++++++++++++++++++++++++++++++++++--->'; print DUMPFILE "\n";
    print DUMPFILE "<h4>Security Policy</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th><th>Value</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>Apply Gateway Rules to Interface Direction</td><td>&nbsp;</td><td>$prop_setting{'gatewaydir'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>TCP Session Timeout (sec)</td><td>&nbsp;</td><td>$prop_setting{'tcptimeout'}</td></tr>\n";    
    print DUMPFILE "<tr valign=top><td>Accept Firewall-1 Control Connections</td><td>$prop_setting{'fw1enable'}</td><td>$prop_setting{'fw1enable_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept UDP Replies</td><td>&nbsp;</td><td>$prop_setting{'udpreply'}</td></td></tr>\n";
    print DUMPFILE "<tr valign=top><td>UDP Reply Timeout (sec)</td><td>&nbsp;</td><td>$prop_setting{'udptimeout'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept Outgoing Packets</td><td>$prop_setting{'outgoing'}</td><td>$prop_setting{'outgoing_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable Decryption on Accept</td><td>&nbsp;</td><td>$prop_setting{'acceptdecrypt'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Use FASTPATH</td><td>&nbsp;</td><td>$prop_setting{'enable_fastpath'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept RIP</td><td>$prop_setting{'rip'}</td><td>$prop_setting{'rip_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept Domain Name Queries (UDP)</td><td>$prop_setting{'domain_udp'}</td><td>$prop_setting{'domain_udp_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept Domain Name Download (TCP)</td><td>$prop_setting{'domain_tcp'}</td><td>$prop_setting{'domain_tcp_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept ICMP</td><td>$prop_setting{'icmpenable'}</td><td>$prop_setting{'icmpenable_p'}</td></tr>\n";
    print DUMPFILE "</table>\n";       
    print DUMPFILE "<h4>Services</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th><th>Value</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable FTP PASV Connections:</td><td>$prop_setting{'ftppasv'}</td><td>$prop_setting{'ftppasv_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable RSH/REXEC Reverse stderr Connections</td><td>$prop_setting{'rshstderr'}</td><td>$prop_setting{'rshstderr_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable RPC Control</td><td>$prop_setting{'rpcenable'}</td><td>$prop_setting{'rpcenable_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable Response of FTP Data Connections:</td><td>$prop_setting{'ftpdata'}</td><td>$prop_setting{'ftpdata_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable Real Audio Reverse Connections:</td><td>$prop_setting{'raudioenable'}</td><td>$prop_setting{'raudioenable_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable VDOLive Reverse Connections</td><td>$prop_setting{'vdolivenable'}</td><td>$prop_setting{'vdolivenable_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable CoolTalk Data Connections (UDP)</td><td>$prop_setting{'cooltalkenable'}</td><td>&nbsp;</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable H.323 Control and Data Connections</td><td>$prop_setting{'iphoneenable'}</td><td>&nbsp;</td></tr>\n";
    print DUMPFILE "</table>\n";
    print DUMPFILE "<h4>Log and Alert</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>Excessive Log Grace Period (sec)</td><td>$prop_setting{'loggrace'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>PopUp Alert Command</td><td>$prop_setting{'alertcmd'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Mail Alert Command</td><td>$prop_setting{'mailcmd'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>SNMP Trap Alert Command</td><td>$prop_setting{'snmptrapcmd'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>User Defined Alert Command</td><td>$prop_setting{'useralertcmd'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Anti Spoof Alert Command</td><td>$prop_setting{'spoofalertcmd'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>User Authentication Alert Command</td><td>$prop_setting{'userauthalertcmd'}</td></tr>\n";    
    print DUMPFILE "<tr valign=top><td>Log Established TCP Packets</td><td>$prop_setting{'log_established_tcp'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Enable Active Connections</td><td>$prop_setting{'liveconns'}</td></tr>\n";
    print DUMPFILE "</table>\n";
    print DUMPFILE "<h4>Resolving</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>Lookup Priorities</td><td>1. $prop_setting{'resolver_1'}<br>2. $prop_setting{'resolver_2'}<br>3. $prop_setting{'resolver_3'}<br>4. $prop_setting{'resolver_4'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>BIND Timeout (sec)</td><td>$prop_setting{'timeout'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>BIND Retries</td><td>$prop_setting{'retries'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Log Viewer Resolver Properties</td><td>$prop_setting{'pagetimeout'}</td></tr>\n";
    print DUMPFILE "</table>\n";
    print DUMPFILE "<h4>Security Servers</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>Telnet Welcome Message File</td><td>$prop_setting{'telnet_msg'}&nbsp;</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Rlogin Welcome Message File</td><td>$prop_setting{'rlogin_msg'}&nbsp;</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>FTP Welcome Message File</td><td>$prop_setting{'ftp_msg'}&nbsp;</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Client Authentication Welcome Message File</td><td>$prop_setting{'clnt_auth_msg'}&nbsp;</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>HTTP Next Proxy</td><td>$prop_setting{'http_next_proxy_host'} : $prop_setting{'http_next_proxy_port'}</td></tr>\n";
    print DUMPFILE "</table>\n";
    print DUMPFILE "<h4>Authentication</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>User Authentication Session Timeout (min)</td><td>$prop_setting{'au_timeout'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>AXENT Pathways Defender Server Setup</td><td>Server IP: $prop_setting{'snk_server_ip'}<br>Agent ID: $prop_setting{'snk_agent_id'}<br>Agent Key: $prop_setting{'snk_agent_key'}</table>\n";
    print DUMPFILE "<h4>SYNDefender</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>Method</td><td>$SynDefender[$prop_setting{'fwsynatk_method'}]</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Timeout</td><td>$prop_setting{'fwsynatk_timeout'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Maximum Sessions</td><td>$prop_setting{'fwsynatk_max'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Display Warning Messages</td><td>$prop_setting{'fwsynatk_warning'}</td></tr>\n";
    print DUMPFILE "</table>\n";
    print DUMPFILE "<h4>Miscellaneous</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>Load Agents Port</td><td>$prop_setting{'load_service_port'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Load Measurement Interval</td><td>$prop_setting{'lbalanced_period_wakeup_sec'}</td></tr>\n";
    print DUMPFILE "</table>\n";
    print DUMPFILE "<h4>Access Lists</h4><table border=2 cellpadding=5 bgcolor=Silver><tr valign=top><th>Property</th><th>Setting</th><th>Value</th></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept Established TCP Connections</td><td>$prop_setting{'established'}</td><td>$prop_setting{'established_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept RIP</td><td>$prop_setting{'rip'}</td><td>$prop_setting{'rip_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept Domain Name Queries (UDP)</td><td>$prop_setting{'domain_udp'}</td><td>$prop_setting{'domain_udp_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept Domain Name Download (TCP)</td><td>$prop_setting{'domain_tcp'}</td><td>$prop_setting{'domain_tcp_p'}</td></tr>\n";
    print DUMPFILE "<tr valign=top><td>Accept ICMP</td><td>$prop_setting{'icmpenable'}</td><td>$prop_setting{'icmpenable_p'}</td></tr>\n";
    print DUMPFILE "</table>\n";
    print DUMPFILE '<!--------------------------------------------------->'; print DUMPFILE "\n";
    print DUMPFILE ''; print DUMPFILE "\n";
    print DUMPFILE ''; print DUMPFILE "\n";
    print DUMPFILE '<!------------------------>'; print DUMPFILE "\n";
    print DUMPFILE '<p><hr>'; print DUMPFILE "\n";
    $now = localtime();
    print DUMPFILE "Generated: $now by $SCR_NAME \($VERSION\)\n";
    print DUMPFILE '</body>'; print DUMPFILE "\n";
    print DUMPFILE '</html>'; print DUMPFILE "\n";
    close (DUMPFILE);
}


##########################################################################
### filling template stuff					       ###
##########################################################################

##########################################################################
# print implicit rules into template
#	file handle
#	location id ( first / "before last" / last )
#	template to be filled
sub subImplicit__Replace {
    my ($OUTFILE)      = $_[0];
    my ($locstr)       = $_[1];
    my ($safeline)     = $_[2];
    my ($sline)        = '';
    my ($line)         = '';

    $safeline =~ s/action/accept/g; 
    $safeline =~ s/track//g;
    $safeline =~ s/time/Any/g;		
    $safeline =~ s/installon/Gateways/g;
    $line = $safeline;
    if ( ("$prop_setting{'fw1enable'}" eq 'true') &&
         (lc("$prop_setting{'fw1enable_p'}") eq "$locstr") ) {
	$sline = $safeline;
	$sline =~ s/comment/Implicit rule: Enable FW1/g;	
	# FW1
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/FW1 Host/g;
	$line =~ s/service/FW1/g;	
	print $OUTFILE "$line\n";
	# FW1_log
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/FW1 Host/g;
	$line =~ s/service/FW1_log/g;
	print $OUTFILE "$line\n";
	# FW1_mgmt
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/gui_clients/g;
	$line =~ s/to/FW1 Management/g;
	$line =~ s/service/FW1_mgmt/g;	
	print $OUTFILE "$line\n";
	# FW1_ela
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FloodGate-1 Host/g;
	$line =~ s/to/FW1 Management/g;
	$line =~ s/service/FW1_ela/g;	
	print $OUTFILE "$line\n";
	# FW1_topo
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/Any/g;
	$line =~ s/to/FW1 Host/g;
	$line =~ s/service/FW1_topo/g;	
	print $OUTFILE "$line\n";
	# FW1_key
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/Any/g;
	$line =~ s/to/FW1 Host/g;
	$line =~ s/service/FW1_key/g;
	print $OUTFILE "$line\n";
	# IKE in
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/Any/g;
	$line =~ s/to/FW1 Host/g;
	$line =~ s/service/IKE/g;
	print $OUTFILE "$line\n";
	# IKE out
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/Any/g;
	$line =~ s/service/IKE/g;
	print $OUTFILE "$line\n";
	# RDP
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/Any/g;
	$line =~ s/to/Any/g;
	$line =~ s/service/RDP/g;	
	print $OUTFILE "$line\n";
	# CVP-Servers
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/CVP-Servers/g;
	$line =~ s/service/FW1_cvp/g;
	print $OUTFILE "$line\n";
	# UFP-Servers
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/UFP-Servers/g;
	$line =~ s/service/FW1_ufp/g;
	print $OUTFILE "$line\n";
	# Radius-Servers
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/Radius-Servers/g;
	$line =~ s/service/RADIUS/g;	
	print $OUTFILE "$line\n";
	# Tacacs-Servers
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/Tacacs-Servers/g;
	$line =~ s/service/TACACS/g;	
	print $OUTFILE "$line\n";
	# Ldap-Servers
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/Ldap-Servers/g;
	$line =~ s/service/ldap/g;	
	print $OUTFILE "$line\n";
	# Logical-Servers
	$line = $sline;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Host/g;
	$line =~ s/to/Logical-Servers/g;
	$line =~ s/service/load_agent/g;	
	print $OUTFILE "$line\n";
    }
# outgoing
    if ( ("$prop_setting{'outgoing'}" eq 'true') &&
         (lc("$prop_setting{'outgoing_p'}") eq "$locstr") ) {
	$line = $safeline;
	$line =~ s/comment/Implicit rule: outgoing/g;	
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/FW1 Module/g;
	$line =~ s/to/Any/g;
	$line =~ s/service/Any/g;	
	print $OUTFILE "$line\n";
    }
# RIP
    if ( ("$prop_setting{'rip'}" eq 'true') &&
         (lc("$prop_setting{'rip_p'}") eq "$locstr") ) {
	$line = $safeline;
	$line =~ s/comment/Implicit rule: RIP/g;	
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/Any/g;
	$line =~ s/to/Any/g;
	$line =~ s/service/RIP/g;	
	print $OUTFILE "$line\n";
    }
# ICMP
    if ( ("$prop_setting{'icmpenable'}" eq 'true') &&
         (lc("$prop_setting{'icmpenable_p'}") eq "$locstr") ) {
	$line = $safeline;
	$line =~ s/comment/Implicit rule: ICMP/g;	
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/Any/g;
	$line =~ s/to/Any/g;
	$line =~ s/service/icmp/g;
	print $OUTFILE "$line\n"; 
    }
# domain TCP
    if ( ("$prop_setting{'domain_tcp'}" eq 'true') &&
         (lc("$prop_setting{'domain_tcp_p'}") eq "$locstr") ) {
	$line = $safeline;
	$line =~ s/comment/Implicit rule: Domain TCP/g;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/Any/g;
	$line =~ s/to/Any/g;
	$line =~ s/service/domain_tcp/g;
	print $OUTFILE "$line\n"; 
    }
# domain UDP
    if ( ("$prop_setting{'domain_udp'}" eq 'true') &&
         (lc("$prop_setting{'domain_udp_p'}") eq "$locstr") ) {
	$line = $safeline;
	$line =~ s/comment/Implicit rule: Domain UDP/g;
	$line =~ s/rule/(implicit)/g;
	$line =~ s/from/Any/g;
	$line =~ s/to/Any/g;
	$line =~ s/service/domain_udp/g;
	print $OUTFILE "$line\n";
    }
}



##########################################################################
# fill template with Rules
#	output file handle
#	line to noodle all the rules through

sub ReplaceTemplateRules{
    my ($OUTFILE)	= $_[0];
    my ($safeline)	= $_[1];
    my ($line)		= '';
    my ($i)		= 0;
    my ($rulenum)	= 0;
    my ($rulenegated)	= '';
    my ($ruleenabled)	= '';
    my ($txt);

    $line = "$safeline";
    for ( $i = 0; $i<=$access_number; $i++ ) {
	if ( $i == 0 ){
	    subImplicit__Replace ($OUTFILE, 'first', "$safeline");
	} elsif ( $i == $access_number ){
	    subImplicit__Replace ($OUTFILE, '"before last"', "$safeline");
	}
	$rulenum = $i + 1;
	if ( $access_disabled[$i] ) { $ruleenabled = "$TemplateITEMSEP (disabled) "; } else { $ruleenabled = ''; }
	$line =~ s/rule/$rulenum$ruleenabled/g;

	if ( $access_from_negated[$i] ) { $rulenegated = ' (negated) '; } else { $rulenegated = ''; }
	$txt = "$access_from[$i]"; $txt =~ s//$TemplateITEMSEP/g;
	$line =~ s/from/$rulenegated$txt/g;
	
	if ( $access_to_negated[$i] ) { $rulenegated = ' (negated) '; } else { $rulenegated = ''; }
	$txt = "$access_to[$i]"; $txt =~ s//$TemplateITEMSEP/g;
	$line =~ s/to/$rulenegated$txt/g;
	
	if ( $access_services_negated[$i] ) { $rulenegated = ' (negated) '; } else { $rulenegated = ''; }
	$txt = "$access_services[$i]"; $txt =~ s//$TemplateITEMSEP/g;
	$line =~ s/service/$rulenegated$txt/g;
	
	$line =~ s/action/$access_action[$i]/g;
	$line =~ s/track/$access_track[$i]/g;
	$line =~ s/time/$access_time[$i]/g;
	$line =~ s/installon/$access_install_on[$i]/g;
	$line =~ s/comment/$access_comment[$i]/g;
	print $OUTFILE "$line\n";
	$line = "$safeline";
    }
    subImplicit__Replace ($OUTFILE, 'last', "$safeline");
}


##########################################################################
# fill template with NAT
#	output file handle
#	line to noodle all the NAT rules through

sub ReplaceTemplateNat{
    my ($OUTFILE)	= $_[0];
    my ($safeline)	= $_[1];
    my ($line)		= '';
    my ($i);
    my ($iplus1);

    #--- first the explicit ones ---
    for ( $i = 0; $i<$nat_number; $i++ ) {
	$line = $safeline;
	$iplus1 = $i + 1;
	$line =~ s/rule/$iplus1/g;
	$line =~ s/installon/$nat_install_on[$i]/g;
	$line =~ s/from/$nat_orig_from[$i]/g;
	$line =~ s/to/$nat_orig_to[$i]/g;
	$line =~ s/service/$nat_orig_svc[$i]/g;
	$line =~ s/tfrom/$nat_transl_from[$i] ($NATtranslation{$nat_transl_from_methd[$i]})/g;
	$line =~ s/tto/$nat_transl_to[$i] ($NATtranslation{$nat_transl_to_methd[$i]})/g;
	$line =~ s/tservice/$nat_transl_svc[$i] ($NATtranslation{$nat_transl_svc_methd[$i]})/g;
	$line =~ s/comment/$nat_comment[$i]/g;
	print $OUTFILE "$line\n";
    }
    #--- now the implicit ones ---
    $safeline =~ s/comment/inplicit NAT set in object definition/g;
    $safeline =~ s/installon/Gateways/g;
    $safeline =~ s/service/Any/g;
    $safeline =~ s/tservice/Original/g;
    if ( $FLAG_implicitrules ) {
	foreach $name (@obj_name){
	    if ( ( $obj_used{$name} || $FLAG_allobjs ) &&
	         ( "$obj_NATadr{$name}" ne '' ) ) {
		#--- forward rule
	        $i++;
		$line = $safeline;
		$line =~ s/rule/$i (implicit)/g;
		$line =~ s/from/$name/g;
		$line =~ s/to/Any/g;
		$line =~ s/tfrom/$obj_NATadr{$name} ($NATtranslation{$obj_NATtype{$name}})/g;
		$line =~ s/tto/Original/g;
		print $OUTFILE "$line\n";
		
		#--- backward rule
	        $i++;
		$line = $safeline;
		$line =~ s/rule/$i (implicit)/g;
		$line =~ s/from/Any/g;
		$line =~ s/to/$obj_NATadr{$name}/g;
		$line =~ s/tfrom/Original/g;
		$line =~ s/tto/$name ($NATtranslation{$obj_NATtype{$name}})/g;
		print $OUTFILE "$line\n";		
	    }	
	}
    }
}


##########################################################################
# fill template with users
#	output file handle
#	line to noodle all the objects through

sub ReplaceTemplateUsers{
    my ($OUTFILE)	= $_[0];
    my ($safeline)	= $_[1];
    my ($line)		= '';
    
    $line = "$safeline";
    foreach $name (@user_name){
	next if($name =~ /^ *$/);
	next if($name =~ /All Users/);
	next if($name =~ /ALL_GROUPS/);
	next if($name =~ /ALL_KEYH/);
	next if($name =~ /ALL_TEMPLATES/);
	next if($user_type{$name} =~ /keyh/);
	next if($user_type{$name} =~ /template/);
	if ( $user_used{$name} || $FLAG_alluser ) {
	   $line =~ s/name/$name/g;
	   $line =~ s/user_type/$user_type{$name}/g;
	   $line =~ s/user_from/$user_from{$name}/g;
	   $line =~ s/user_to/$user_to{$name}/g;
	   $line =~ s/user_auth/$user_auth{$name}/g;
	   $line =~ s/user_day/$user_day{$name}/g;
	   $line =~ s/user_expires/$user_expires{$name}/g;
	   $txt = "$user_members{$name}"; $txt =~ s//$TemplateITEMSEP/g;
	   $line =~ s/user_members/$txt/g;
	   $line =~ s/user_comment/$user_comment{$name}/g;
	print $OUTFILE "$line\n";
	$line = "$safeline";
      }
    }
}


##########################################################################
# fill template with properties
#	output file handle
#	line to noodle all the objects through

sub ReplaceTemplateProperties{
    my ($OUTFILE)	= $_[0];
    my ($safeline)	= $_[1];
    my ($line)		= '';
    
    $line = "$safeline";
    
    $line =~ s/include/$TemplateINCLUDE/g;
    
    $line =~ s/prop_gatewaydir/$prop_setting{'gatewaydir'}/g;
    $line =~ s/prop_tcptimeout/$prop_setting{'tcptimeout'}/g;
    $line =~ s/prop_fw1enable/$prop_setting{'fw1enable'}/g;
    $line =~ s/prop_fw1enable_p/$prop_setting{'fw1enable_p'}/g;
    $line =~ s/prop_udpreply/$prop_setting{'udpreply'}/g;
    $line =~ s/prop_udpreply_p/$prop_setting{'udpreply_p'}/g;
    $line =~ s/prop_udptimeout/$prop_setting{'udptimeout'}/g;
    $line =~ s/prop_outgoing/$prop_setting{'outgoing'}/g;
    $line =~ s/prop_outgoing_p/$prop_setting{'outgoing_p'}/g;
    $line =~ s/prop_acceptdecrypt/$prop_setting{'acceptdecrypt'}/g;
    $line =~ s/prop_enable_fastpath/$prop_setting{'enable_fastpath'}/g;
    $line =~ s/prop_rip/$prop_setting{'rip'}/g;
    $line =~ s/prop_rip_p/$prop_setting{'rip_p'}/g;
    $line =~ s/prop_domain_udp/$prop_setting{'domain_udp'}/g;
    $line =~ s/prop_domain_udp_p/$prop_setting{'domain_udp_p'}/g;
    $line =~ s/prop_domain_tcp/$prop_setting{'domain_tcp'}/g;
    $line =~ s/prop_domain_tcp_p/$prop_setting{'domain_tcp_p'}/g;
    $line =~ s/prop_icmpenable/$prop_setting{'icmpenable'}/g;
    $line =~ s/prop_icmpenable_p/$prop_setting{'icmpenable_p'}/g;
    $line =~ s/prop_ftppasv/$prop_setting{'ftppasv'}/g;
    $line =~ s/prop_ftppasv_p/$prop_setting{'ftppasv_p'}/g;
    $line =~ s/prop_rshstderr/$prop_setting{'rshstderr'}/g;
    $line =~ s/prop_rshstderr_p/$prop_setting{'rshstderr_p'}/g;
    $line =~ s/prop_rpcenable/$prop_setting{'rpcenable'}/g;
    $line =~ s/prop_rpcenable_p/$prop_setting{'rpcenable_p'}/g;
    $line =~ s/prop_ftpdata/$prop_setting{'ftpdata'}/g;
    $line =~ s/prop_ftpdata_p/$prop_setting{'ftpdata_p'}/g;
    $line =~ s/prop_raudioenable/$prop_setting{'raudioenable'}/g;
    $line =~ s/prop_raudioenable_p/$prop_setting{'raudioenable_p'}/g;
    $line =~ s/prop_vdolivenable/$prop_setting{'vdolivenable'}/g;
    $line =~ s/prop_vdolivenable_p/$prop_setting{'vdolivenable_p'}/g;
    $line =~ s/prop_cooltalkenable/$prop_setting{'cooltalkenable'}/g;
    $line =~ s/prop_iphoneenable/$prop_setting{'iphoneenable'}/g;
    $line =~ s/prop_loggrace/$prop_setting{'loggrace'}/g;
    $line =~ s/prop_alertcmd/$prop_setting{'alertcmd'}/g;
    $line =~ s/prop_mailcmd/$prop_setting{'mailcmd'}/g;
    $line =~ s/prop_snmptrapcmd/$prop_setting{'snmptrapcmd'}/g;
    $line =~ s/prop_useralertcmd/$prop_setting{'useralertcmd'}/g;
    $line =~ s/prop_spoofalertcmd/$prop_setting{'spoofalertcmd'}/g;
    $line =~ s/prop_userauthalertcmd/$prop_setting{'userauthalertcmd'}/g;
    $line =~ s/prop_log_established_tcp/$prop_setting{'log_established_tcp'}/g;
    $line =~ s/prop_liveconns/$prop_setting{'liveconns'}/g;
    $line =~ s/prop_resolver_1/$prop_setting{'resolver_1'}/g;
    $line =~ s/prop_resolver_2/$prop_setting{'resolver_2'}/g;
    $line =~ s/prop_resolver_3/$prop_setting{'resolver_3'}/g;
    $line =~ s/prop_resolver_4/$prop_setting{'resolver_4'}/g;
    $line =~ s/prop_timeout/$prop_setting{'timeout'}/g;
    $line =~ s/prop_retries/$prop_setting{'retries'}/g;
    $line =~ s/prop_pagetimeout/$prop_setting{'pagetimeout'}/g;
    $line =~ s/prop_telnet_msg/$prop_setting{'telnet_msg'}/g;
    $line =~ s/prop_rlogin_msg/$prop_setting{'rlogin_msg'}/g;
    $line =~ s/prop_ftp_msg/$prop_setting{'ftp_msg'}/g;
    $line =~ s/prop_clnt_auth_msg/$prop_setting{'clnt_auth_msg'}/g;
    $line =~ s/prop_http_next_proxy_host/$prop_setting{'http_next_proxy_host'}/g;
    $line =~ s/prop_http_next_proxy_port/$prop_setting{'http_next_proxy_port'}/g;
    $line =~ s/prop_au_timeout/$prop_setting{'au_timeout'}/g;
    $line =~ s/prop_snk_server_ip/$prop_setting{'snk_server_ip'}/g;
    $line =~ s/prop_snk_agent_id/$prop_setting{'snk_agent_id'}/g;
    $line =~ s/prop_snk_agent_key/$prop_setting{'snk_agent_key'}/g;
    $line =~ s/prop_fwsynatk_method/$prop_setting{'fwsynatk_method'}/g;
    $line =~ s/prop_fwsynatk_timeout/$prop_setting{'fwsynatk_timeout'}/g;
    $line =~ s/prop_fwsynatk_max/$prop_setting{'fwsynatk_max'}/g;
    $line =~ s/prop_fwsynatk_warning/$prop_setting{'fwsynatk_warning'}/g;
    $line =~ s/prop_load_service_port/$prop_setting{'load_service_port'}/g;
    $line =~ s/prop_lbalanced_period_wakeup_sec/$prop_setting{'lbalanced_period_wakeup_sec'}/g;
    $line =~ s/prop_established/$prop_setting{'established'}/g;
    $line =~ s/prop_established_p/$prop_setting{'established_p'}/g;
    $line =~ s/prop_rip/$prop_setting{'rip'}/g;
    $line =~ s/prop_rip_p/$prop_setting{'rip_p'}/g;
    $line =~ s/prop_domain_udp/$prop_setting{'domain_udp'}/g;
    $line =~ s/prop_domain_udp_p/$prop_setting{'domain_udp_p'}/g;
    $line =~ s/prop_domain_tcp/$prop_setting{'domain_tcp'}/g;
    $line =~ s/prop_domain_tcp_p/$prop_setting{'domain_tcp_p'}/g;
    $line =~ s/prop_icmpenable/$prop_setting{'icmpenable'}/g;
    $line =~ s/prop_icmpenable_p/$prop_setting{'icmpenable_p'}/g;
    print $OUTFILE "$line\n";
}


##########################################################################
# fill template with objects
#	output file handle
#	line to noodle all the objects through

sub ReplaceTemplateObjects{
    my ($OUTFILE)	= $_[0];
    my ($safeline)	= $_[1];
    my ($line)		= '';
    my ($nattype)	= '';
    my ($loc)		= '';
    my ($fw1)		= '';
    
    $line = "$safeline";
    foreach $name (@obj_name){
      if ( $obj_used{$name} || $FLAG_allobjects ) {
	$line =~ s/name/$name/g;
	$line =~ s/type/$obj_type{$name}/g;
	if ( $obj_location{$name} ) { $loc = 'external'; } else { $loc = 'internal'; } 	
	$line =~ s/location/$loc/g;
	if ( $obj_is_fw1{$name} ) { $fw1 = 'FW1 installed'; } else { $fw1 = '---'; } 	
	$line =~ s/fw1/$fw1/g;
	$line =~ s/ipaddress/$obj_ipaddr{$name}/g;
	$line =~ s/netmask/$obj_netmask{$name}/g;
	$line =~ s/nataddress/$obj_NATadr{$name}/g;
	$line =~ s/nattype/$NATtranslation{$obj_NATtype{$name}}/g;
	$txt = "$obj_members{$name}"; $txt =~ s//$TemplateITEMSEP/g;
	$line =~ s/members/$txt/g;
	$line =~ s/comment/$obj_comment{$name}/g;
	print $OUTFILE "$line\n";
	$line = "$safeline";
      }
    }
}


##########################################################################
# fill template with services
#	output file handle
#	line to noodle all the objects through

sub ReplaceTemplateServices{
    my ($OUTFILE)	= $_[0];
    my ($safeline)	= $_[1];
    my ($line)		= '';
    my ($svcmem)	= '';
    
    $line = "$safeline";
    foreach $name (@svc_name){
      if ( $svc_used{$name} || $FLAG_allservices ) {
	$line =~ s/name/$name/g;
	$line =~ s/type/$svc_type{$name}/g;
	$line =~ s/port/$svc_dst_port{$name}/g;
	$line =~ s/sportfrom/$svc_low{$name}/g;
	$line =~ s/sportto/$svc_high{$name}/g;
	$line =~ s/match/$svc_match{$name}/g;
	$line =~ s/prolog/$svc_prolog{$name}/g;
	$svcmem = "$svc_members{$name}"; $svcmem =~ s//$TemplateITEMSEP/g;
	$line =~ s/members/$svcmem/g;
	$line =~ s/comment/$svc_comment{$name}/g;
	print $OUTFILE "$line\n";
	$line = "$safeline";
      }
    }
}


##########################################################################
##########################################################################
# fill template and replace with current config
#	filename of the template
#	filename of the output file

sub Fill_Template {
    my ($template_filename)	= $_[0];
    my ($output_filename)	= $_[1];
    my ($dummy)			= '';
    my ($line)			= '';
    my ($safeline)		= '';

    open (INFILE,"<$template_filename") 
    	or die "Can't read template $template_filename!\n";
    open (DUMPFILE,">$output_filename")
    	or die "Can't write template output file $output_filename!\n";
    $line = <INFILE>;
    if ( "$line" !~ m/^--------------*/ ) { die "Template has broken header (line 1)\n"; }
    $line = <INFILE>;
    ($dummy,$TemplateITEMSEP) = split(/=/, &fromdos($line), 2);
    &PrintLog("Template: $TemplateITEMSEP\n");
    if ( "$dummy" ne 'title'  ) { die "Template has broken header (line 2 not title)\n"; }
    $line = <INFILE>;
    ($dummy,$TemplateITEMSEP) = split(/=/, &fromdos($line), 2);
    $TemplateITEMSEP =~ s/<NEWLINE>/\n/g;
    $TemplateITEMSEP =~ s/<TAB>/\t/g;
    $TemplateITEMSEP =~ s/<SPACE>/ /g;
    if ( "$dummy" ne 'separator'  ) { die "Template has broken header (line 3 not separator)\n"; }
    $line = <INFILE>;
    if ( "$line" !~ m/^--------------*/ ) { die "Template has broken header (line 4)\n"; }
    while ( "$line" ne '' ) {
        while ( ($line = <INFILE>) && ($line !~ m/.*/x)){
		$line =~ s/\n//g; $line =~ s/\r//g;
		$now = localtime();
		$line =~ s/configset/$FW1rules/g;
		$line =~ s/date/$now/g;
		$line =~ s/fwrules/$SCR_NAME \($VERSION\)/g;
		$line =~ s/<NEWLINE>/\n/g;
		$line =~ s/<TAB>/\t/g;
		$line =~ s/<SPACE>/ /g;
		&ReplaceTemplateProperties(DUMPFILE,$line);
        }
        # ------------ read loopline --------------
        $safeline = &fromdos("$line");
        while ( ($line = <INFILE>) && ($line =~ m/.*/x)){
		$line =~ s/\n//g; $line =~ s/\r//g;
                $safeline = "$safeline\n$line";
		$line = '';
        }	
	$line =~ s/<NEWLINE>/\n/g;
	$line =~ s/<TAB>/\t/g;
	$line =~ s/<SPACE>/ /g;
	if ( $safeline =~ m/objects.*/ ) {
	   $safeline =~ s/objects//g;
	   if ( ! $FLAG_noobjs ) {
		&ReplaceTemplateObjects(DUMPFILE,$safeline);
	   }
	}
	if ( $safeline =~ m/rules.*/ ) {
	   $safeline =~ s/rules//g;
	   if ( ! $FLAG_norules ) {
		&ReplaceTemplateRules(DUMPFILE,$safeline);
	   }
	}
	if ( $safeline =~ m/services.*/ ) {
	   $safeline =~ s/services//g;
	   if ( ! $FLAG_noservices ) {
		&ReplaceTemplateServices(DUMPFILE,$safeline);
	   }
	}
	if ( $safeline =~ m/nat.*/ ) {
	   $safeline =~ s/nat//g;
	   if ( ! $FLAG_norules ) {
		&ReplaceTemplateNat(DUMPFILE,$safeline);
           }
	}
	if ( $safeline =~ m/user.*/ ) {
	   $safeline =~ s/user//g;
	   if ( ! $FLAG_nousers ) {
		&ReplaceTemplateUsers(DUMPFILE,$safeline);
	   }
	}
	&ReplaceTemplateProperties(DUMPFILE,$line);
    }
}



##########################################################################
##########################################################################
###   MAIN
##########################################################################
##########################################################################

# Parse and process options
if (!GetOptions(\%optctl, 
	'objects=s', 'rules=s', 'users=s',
	'all_objects', 'all_services', 'all_users', 
	'with_implicit_rules', 'with_ip', 
	'no_rules', 'no_objects', 'no_services', 'no_users',
	'with_colors', 'icon_path=s', 'use-css=s', 'title=s', 
	'dump_nat_tsv=s', 'dump_nat_txt=s', 
	'dump_objects_tsv=s', 'dump_objects_txt=s', 
	'dump_rules_tsv=s', 'dump_rules_txt=s', 
	'dump_services_tsv=s', 'dump_services_txt=s',	
	'dump_users_tsv=s', 'dump_users_txt=s',	
	'output_html=s',
	'template=s', 'template_include=s', 'output=s',
	'verbose','sort_by_type',
	'dump_unused_objects=s','dump_properties=s'
	)
	|| keys(%optctl) == 0 || $optctl{help} == 1 || $optctl{version} == 1 )
{
        if ($optctl{version} == 1)
        {
		print STDERR "$SCR_NAME \($VERSION\) - 2001 by Volker Tanger \<volker.tanger\@wyae.de\>\n";
        } else {
                &Usage();
        }
        exit;
}


#--------------------------------------------------
# filename options
if (defined($optctl{'icon_path'})) { $IconPathName = $optctl{'icon_path'}; }
if (defined($optctl{'objects'})) { $FW1objects = $optctl{'objects'}; }
if (defined($optctl{'rules'})) { $FW1rules = $optctl{'rules'}; }
if (defined($optctl{'users'})) { $FW1user = $optctl{'users'}; }

#--------------------------------------------------
# HTML / Template candy
if (defined($optctl{'use-css'})) { $HTMLcssfile = $optctl{'use-css'}; }
if (defined($optctl{'title'})) { $HTMLtitle = $optctl{'title'}; }

if (defined($optctl{'template_include'})) { $TemplateINCLUDE = $optctl{'template_include'}; }

#--------------------------------------------------
# switches / flags
$FLAG_withcolors = (defined($optctl{'with_colors'}));
$FLAG_withip = (defined($optctl{'with_ip'}));
$FLAG_allobjs = (defined($optctl{'all_objects'}));
$FLAG_allservices = (defined($optctl{'all_services'}));
$FLAG_alluser = (defined($optctl{'all_users'}));
$FLAG_implicitrules = (defined($optctl{'with_implicit_rules'}));
$FLAG_verbose = (defined($optctl{'verbose'}));
$FLAG_sortbytype = (defined($optctl{'sort_by_type'}));
$FLAG_norules = (defined($optctl{'no_rules'}));
$FLAG_noobjs = (defined($optctl{'no_objects'}));
$FLAG_noservices = (defined($optctl{'no_services'}));
$FLAG_nousers = (defined($optctl{'no_users'}));

if (! $FLAG_verbose) {
   print "$SCR_NAME \($VERSION\) - 2001 by Volker Tanger \<volker.tanger\@wyae.de\>\n\n";
}   



if ( ( $FLAG_noobjs && $FLAG_allobjs ) ||
     ( $FLAG_noservices && $FLAG_allservices ) ||
     ( $FLAG_nousers && $FLAG_allusers ) ) {
   die "Conflicting  --all_*  and --no_* options. Aborting.";
}

if ( ( $FLAG_noobjs && ( $optctl{'dump_objects_txt'} || $optctl{'dump_objects_tsv'} ) ) ||
     ( $FLAG_noservices && ( $optctl{'dump_services_txt'} || $optctl{'dump_services_tsv'} ) ) ||
     ( $FLAG_nousers && ( $optctl{'dump_users_txt'} || $optctl{'dump_users_tsv'} ) ) ) {
   die "Conflicting  --no_*  and --dump_* options. Aborting.";
}

if ( $FLAG_norules ) {
   $FLAG_allobjs = 1;
   $FLAG_allservices = 1;
   $FLAG_allusers = 1;
}


if ( ( (defined($optctl{'template'})) && 
       (! defined($optctl{'output'})) ) 
       ||
     ( (! defined($optctl{'template'})) && 
       (defined($optctl{'output'})) ) ){
   die "ERROR: --output option needs --template - Aborting.";
}


#----------------------------------------------------------------

open (LOGFILE,">$LogFile") ;

&PrintLog("$SCR_NAME \($VERSION\) - 2001 by Volker Tanger \<volker.tanger\@wyae.de\>\n\n");

#------ first the objects ------

open (INFILE,"$FW1objects") 
	or die "Cannot open the object file $FW1objects!\n\n";
	
&PrintLog("skipping...");
while ( ($line = <INFILE>) && ( $line !~ /^\t\:netobj \(netobj/ ) ) { &PrintLog('.'); }
if ( ! $FLAG_noobjs ) { 
   &PrintLog("\n\nReading network objects...");
   &ReadNetworkObjects; 
}

&PrintLog("\n\nskipping...");
while ( ($line = <INFILE>) && ( $line !~ /^\t\:servobj \(servobj/ ) ) { &PrintLog('.');}
if ( ! $FLAG_noservices ) {
   &PrintLog("\n\nReading services...");
   &ReadServices;
}   

&PrintLog("\n\nskipping...");
while ( ($line = <INFILE>) && ( $line !~ /^\t\:resourcesobj \(resourcesobj/ ) ) { &PrintLog('.');}
&PrintLog("\n\nReading resources...");
#&ReadResources;

&PrintLog("\n\nskipping...");
while ( ($line = <INFILE>) && ( $line !~ /^\t\:props \(/ ) ) { &PrintLog('.');}
&PrintLog("\n\nReading properties...");
&ReadProperties;

close (INFILE);

#------ now the rulebase ------

if ( ! $FLAG_norules ) {

   open (INFILE,"$FW1rules") 
	or die "Cannot open the rules file $FW1rules!\n\n";
	
   &PrintLog("\n\nskipping...");
   while ( ( $line !~ /^\t\:rule \(/ ) && ($line = <INFILE>) ) { &PrintLog('.');}
   &PrintLog("\n\nReading access rules...");
   &ReadAccessRules;

   &PrintLog("\n\nskipping...");
   while ( ( $line !~ /^\t\:rule_adtr \(/ ) && ($line = <INFILE>) ) { &PrintLog('.');}
   &PrintLog("\n\nReading NAT rules...");
   &ReadNATrules;

   close (INFILE);
}

#------ now the user ------

if ( ! $FLAG_nousers ) {
   &PrintLog("\n\nReading Users...");
   &ReadUser;
}
&PrintLog("\n\nReading Done.\n\n");

#--------------------------------------------------
# dump data into separate files
if (defined($optctl{'dump_nat_tsv'})) {
	&PrintLog("Dumping NAT in TSV format.\n");
	&DumpNatTSV ($optctl{'dump_nat_tsv'}); }
if (defined($optctl{'dump_nat_txt'})) { 
	&PrintLog("Dumping NAT in TXT format.\n");
	&DumpNatTXT ($optctl{'dump_nat_txt'}); }

if (defined($optctl{'dump_rules_tsv'})) {
	&PrintLog("Dumping Rules in TSV format.\n");
	&DumpRulesTSV ($optctl{'dump_rules_tsv'}); }
if (defined($optctl{'dump_rules_txt'})) { 
	&PrintLog("Dumping Rules in TXT format.\n");
	&DumpRulesTXT ($optctl{'dump_rules_txt'}); }

if (defined($optctl{'dump_objects_tsv'})) {
	&PrintLog("Dumping Objects in TSV format.\n");
	&DumpObjectsTSV ($optctl{'dump_objects_tsv'}); }
if (defined($optctl{'dump_objects_txt'})) { 
	&PrintLog("Dumping Objects in TXT format.\n");
	&DumpObjectsTXT ($optctl{'dump_objects_txt'}); }

if (defined($optctl{'dump_services_tsv'})) {
	&PrintLog("Dumping Services in TSV format.\n");
	&DumpServicesTSV ($optctl{'dump_services_tsv'}); }
if (defined($optctl{'dump_services_txt'})) { 
	&PrintLog("Dumping Services in TXT format.\n");
	&DumpServicesTXT ($optctl{'dump_services_txt'}); }

if (defined($optctl{'dump_users_tsv'})) {
	&PrintLog("Dumping Users in TSV format.\n");
	&DumpUsersTSV ($optctl{'dump_users_tsv'}); }
if (defined($optctl{'dump_users_txt'})) {
	&PrintLog("Dumping Users in TXT format.\n");
	&DumpUsersTXT ($optctl{'dump_users_txt'}); }

if (defined($optctl{'dump_unused_objects'})) { 
	&PrintLog("Dumping unused objects in TXT file.\n");
	&DumpUnusedObjects ($optctl{'dump_unused_objects'}); }

if (defined($optctl{'dump_properties'})) { 
	&PrintLog("Dumping properties in TXT file.\n");
	&DumpProperties ($optctl{'dump_properties'}); }

#--------------------------------------------------
# print ruleset
if (defined($optctl{'output_html'})) {
	&PrintLog("Printing ruleset in HTML format.\n");
	&Output_in_HTML ($optctl{'output_html'}); }

if ( (defined($optctl{'template'})) && (defined($optctl{'output'})) ) {
	&PrintLog("Printing ruleset into template.\n");
	&Fill_Template ($optctl{'template'},$optctl{'output'}); }

	&PrintLog("\nDone.\n\n");
close (LOGFILE);

#############################################################################
